/******************************************************************************
CIF_I_Comm_Endpoint.C

Description:

  Classes to provide simple Nexus endpoint functionality

CVS Information:

  $Source: /home/dsl/Master/cif/CIF-Current/cpp/libraries/comm/CIF_I_Comm_Endpoint.C,v $
  $Date: 1998/10/29 19:26:51 $
  $Revision: 1.9 $
  $State: Exp $
  $Author: toonen $
******************************************************************************/

/******************************************************************************
			     Include header files
******************************************************************************/
#include "CIF_I_Comm_Endpoint.H"


/******************************************************************************
			      Initalize constants
******************************************************************************/
const int CIF_I_Comm_Endpoint::i_handle_Message_HID = 0;



/******************************************************************************
			  CIF_I_Comm_Endpoint methods
******************************************************************************/

/*
 * CIF_I_Comm_Endpoint::CIF_I_Comm_Endpoint()
 */
CIF_I_Comm_Endpoint::CIF_I_Comm_Endpoint()
{
    /*
     * Initialize internal state and data structures
     */
    l_q_head = CIF_NULL;
    l_q_tail = CIF_NULL;
    l_blocking_cnt = 0;
    l_deactivate_pending = false;
    l_attach_enabled = false;
    l_attach_port = 0;
    l_attach_hostname = CIF_NULL;
    l_event_state = EVENT_STATE_DISABLED;
    l_event_listener = CIF_NULL;
    l_dispatching_event = false;
    l_object_state = OBJECT_STATE_INACTIVE;

    CIF_Comm::status = CIF_Comm_Status::SUCCESS;    
}
/* CIF_I_Comm_Endpoint::CIF_I_Comm_Endpoint() */


/*
 * CIF_I_Comm_Endpoint::activate()
 */
CIF_Bool
CIF_I_Comm_Endpoint::activate(
    ARG_IN const Properties &			properties)
{
    static const globus_nexus_handler_t		handler_table[] =
    {
	{GLOBUS_NEXUS_HANDLER_TYPE_NON_THREADED,
	 l_handle_Message},
	{GLOBUS_NEXUS_HANDLER_TYPE_NON_THREADED,
	 (globus_nexus_handler_func_t) GLOBUS_NULL},
    };
    
    CIF_Toolkit_Monitor				monitor(l_mutex);

    CIF_Comm_Status				status;
    
    /*
     * Verify that the object is in the correct state
     */
    status = l_verify_Object_State(OBJECT_STATE_INACTIVE);
    if (status != CIF_Comm_Status::SUCCESS)
    {
	CIF_Comm::status = status;
	return false;
    }
    
    /*
     * Create a endpoint attribute using the provided information
     */
    globus_nexus_endpointattr_t			ep_attr;
    
    globus_nexus_endpointattr_init(&ep_attr);

    globus_nexus_endpointattr_set_handler_table(
	&ep_attr,
	(globus_nexus_handler_t *) handler_table,
	(sizeof(handler_table) / sizeof(globus_nexus_handler_t)) - 1);

    globus_nexus_endpointattr_set_unknown_handler(
	&ep_attr,
	l_handle_Unknown,
	GLOBUS_NEXUS_HANDLER_TYPE_NON_THREADED);

    /*
     * If the user requested a specific protocol to use, then attempt to use
     * that protocol.
     */
    int						rc;
    globus_nexus_proto_type_t			proto_type;
    void *					proto_info_ptr;
    int						proto_info_size;

    rc = GLOBUS_SUCCESS;

    properties.get_Protocol(proto_type,
			    proto_info_ptr,
			    proto_info_size);
    
    if (proto_type != GLOBUS_NEXUS_PROTO_TYPE_ALL)
    {
	rc = globus_nexus_endpointattr_set_protocol(&ep_attr,
						    proto_type,
						    proto_info_ptr,
						    proto_info_size);
	if (rc != GLOBUS_SUCCESS)
	{
	    goto fn_abort;
	}
    }
    
    /*
     * Initialize the internal endpoint
     */
    rc = globus_nexus_endpoint_init(&l_nx_ep, &ep_attr);    
    if (rc != GLOBUS_SUCCESS)
    {
	goto fn_abort;
    }

    globus_nexus_endpoint_set_user_pointer(&l_nx_ep, (void *) this);


    /*
     * If specified, enabled remote attachment services
     */
    if (properties.is_Attach_Enabled())
    {
	l_attach_port = properties.get_Attach_Port();
	rc = globus_nexus_allow_attach(&l_attach_port,
				       &l_attach_hostname,
				       l_handle_Attach,
				       (void *) this);
	if (rc != GLOBUS_SUCCESS)
	{
	    globus_nexus_endpoint_destroy(&l_nx_ep);
	    goto fn_abort;
	}
	
	l_attach_enabled = true;
    }
    else
    {
	l_attach_enabled = false;
    }

    /*
     * Set pointer to event listener class which will be handling message
     * arrival events
     */
    l_event_listener = properties.get_Event_Listener();
    
  fn_abort:
    /*
     * Free up resources used by the endpoint attribute
     */
    globus_nexus_endpointattr_destroy(&ep_attr);

    /*
     * Set the internal states and global status, and return
     */
    if (rc == GLOBUS_SUCCESS)
    {
	l_object_state = OBJECT_STATE_ACTIVE;
	CIF_Comm::status = CIF_Comm_Status::SUCCESS;
	return true;
    }
    else
    {
	l_attach_enabled = false;
	l_attach_port = 0;
	l_attach_hostname = CIF_NULL;
	l_event_listener = CIF_NULL;
	

	CIF_Comm::status = CIF_Comm_Status::FAILURE;
	return false;
    }
}
/* CIF_I_Comm_Endpoint::activate() */


/*
 * CIF_I_Comm_Endpoint::receive()
 */
CIF_Bool
CIF_I_Comm_Endpoint::receive(
    ARG_OUT globus_nexus_buffer_t &		buffer,
    ARG_IN const CIF_Toolkit_Timeout &	 	timeout)
{
    CIF_Toolkit_Monitor
	monitor(l_mutex, l_q_cond);

    CIF_Comm_Status				status;
    CIF_Bool					rc;
    
    /*
     * Verify that the object is in the correct state
     */
    status = l_verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != CIF_Comm_Status::SUCCESS)
    {
	CIF_Comm::status = status;
	return false;
    }

    /*
     * Wait for a message to arrive, the timer to expire, or for something to
     * go wrong
     */
    if (l_q_head == CIF_NULL)
    {
	if (l_dispatching_event  &&
	    l_dispatching_thread == globus_thread_self())
	{
	    CIF_Comm::status = CIF_Comm_Status::WOULD_BLOCK;
	    return false;
	}
    
	for (;;)
	{
	    if (l_deactivate_pending)
	    {
		CIF_Comm::status = CIF_Comm_Status::OBJECT_INACTIVE;
		rc = false;
		break;
	    }
	    
	    l_blocking_cnt++;
	    monitor.wait(timeout);
	    l_blocking_cnt--;

	    if (l_q_head != CIF_NULL)
	    {
		rc = true;
		break;
	    }

	    if (timeout == 0)
	    {
		CIF_Comm::status = CIF_Comm_Status::TIMED_OUT;
		rc = false;
		break;
	    }
	}

	if (l_deactivate_pending && l_blocking_cnt == 0)
	{
	    l_deactivate_cond.broadcast();
	}

	if (rc == false)
	{
	    return false;
	}
    }

    /*
     * A new message arrived, so initialize the user's buffer with the message
     *
     * TODO: This code should be fixed once Nexus dispatch is fixed.  We are
     * making the assumption that globus_nexus_buffer_t will fit in a (void *),
     * which it currently will, but may not in the future.
     */
    Queue_Entry *				entry;

    entry = l_q_head;
    buffer = entry->buf;
    l_q_head = entry->next;
    if (l_q_head == CIF_NULL)
    {
	l_q_tail = CIF_NULL;
    }
    delete entry;
    
    CIF_Comm::status = CIF_Comm_Status::SUCCESS;
    return true;
}
/* CIF_I_Comm_Endpoint::receive() */


/*
 * CIF_I_Comm_Endpoint::poll()
 */
CIF_Bool
CIF_I_Comm_Endpoint::poll()
{
    globus_poll();

    {
	CIF_Toolkit_Monitor			monitor(l_mutex);
	CIF_Comm_Status				status;

	status = l_verify_Object_State(OBJECT_STATE_ACTIVE);
	if (status != CIF_Comm_Status::SUCCESS)
	{
	    CIF_Comm::status = status;
	    return false;
	}
	
	CIF_Comm::status = CIF_Comm_Status::SUCCESS;
	return ((l_q_head == CIF_NULL) ? false : true);
    }
}
/* CIF_I_Comm_Endpoint::poll() */


/*
 * CIF_I_Comm_Endpoint::enable_Event_Listener()
 */
CIF_Bool
CIF_I_Comm_Endpoint::enable_Event_Listener()
{
    CIF_Toolkit_Monitor				monitor(l_mutex);
    
    if (l_event_listener == CIF_NULL)
    {
	CIF_Comm::status = CIF_Comm_Status::FAILURE;
	return false;
    }

    if (l_event_state == EVENT_STATE_DISABLED)
    {
	Queue_Entry *				entry;

	l_event_state = EVENT_STATE_ENABLING;
	
	l_blocking_cnt++;

	l_dispatching_event = true;
	l_dispatching_thread = globus_thread_self();
	
	entry = l_q_head;
	while(entry != CIF_NULL)
	{
	    if (entry->event_dispatched == false)
	    {
		entry->event_dispatched = true;
		
		{
		    l_mutex.unlock();
	
		    Message_Arrival_Event	event(this);
		    
		    l_event_listener->
			handle_Event(* (CIF_Toolkit_Event *) &event);

		    l_mutex.lock();
		}
		
		entry = l_q_head;

		if (l_event_state == EVENT_STATE_DISABLED)
		{
		    l_event_cond.broadcast();
		    break;
		}

		if (l_deactivate_pending)
		{
		    break;
		}
	    }
	    else
	    {
		entry = entry->next;
	    }
	}

	l_dispatching_event = false;
	
	l_blocking_cnt--;

	if (l_deactivate_pending && l_blocking_cnt == 0)
	{
	    l_deactivate_cond.broadcast();
	}
    }
    
    if (l_event_state == EVENT_STATE_ENABLING)
    {
	l_event_state = EVENT_STATE_ENABLED;
    }
    
    CIF_Comm::status = CIF_Comm_Status::SUCCESS;
    return true;
}
/* CIF_I_Comm_Endpoint::enable_Event_Listener() */

/*
 * CIF_I_Comm_Endpoint::disable_Event_Listener()
 */
CIF_Bool
CIF_I_Comm_Endpoint::disable_Event_Listener()
{
    CIF_Toolkit_Monitor				monitor(l_mutex);

    if (l_event_listener == CIF_NULL)
    {
	CIF_Comm::status = CIF_Comm_Status::FAILURE;
	return false;
    }

    if (l_event_state != EVENT_STATE_DISABLED)
    {
	l_event_state = EVENT_STATE_DISABLED;

	if (l_dispatching_thread != globus_thread_self())
	{
	    while (l_dispatching_event)
	    {
		l_blocking_cnt++;
		l_event_cond.wait(l_mutex, 0);
		l_blocking_cnt--;
	    }
	}
    }
    
    CIF_Comm::status = CIF_Comm_Status::SUCCESS;
    return true;
}
/* CIF_I_Comm_Endpoint::disable_Event_Listener() */


/*
 * CIF_I_Comm_Endpoint::get_Attach_Hostname()
 */
const char *
CIF_I_Comm_Endpoint::get_Attach_Hostname() const
{
    return l_attach_hostname;
}
/* CIF_I_Comm_Endpoint::get_Attach_Hostname() */


/*
 * CIF_I_Comm_Endpoint::get_Attach_Port()
 */
unsigned short
CIF_I_Comm_Endpoint::get_Attach_Port() const
{
    return l_attach_port;
}
/* CIF_I_Comm_Endpoint::get_Attach_Port() */


/*
 * CIF_I_Comm_Endpoint::l_deactivate()
 */
CIF_Bool
CIF_I_Comm_Endpoint::l_deactivate(
    CIF_Toolkit_Monitor &			monitor)
{
    CIF_Comm_Status				status;
    
    /*
     * Verify that we are in the correct state
     */
    status = l_verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != CIF_Comm_Status::SUCCESS)
    {
	CIF_Comm::status = status;
	return false;
    }

    /*
     * Wait for any blocking receives or dispatched events to return
     */
    l_deactivate_pending = true;

    while (l_blocking_cnt > 0)
    {
	/*
	 * Wakeup any threads waiting for messages to arrive
	 */
	l_q_cond.broadcast();

	/*
	 * Wait until the last blocking thread signals that it has completed
	 */
	monitor.wait();
	
	/*
	 * Verify that we are still in the correct state
	 */
	status = l_verify_Object_State(OBJECT_STATE_ACTIVE);
	if (status != CIF_Comm_Status::SUCCESS)
	{
	    CIF_Comm::status = status;
	    return false;
	}
    }

    l_deactivate_pending = false;

    /*
     * Disable remote attach services
     */
    if (l_attach_enabled)
    {
	globus_nexus_disallow_attach(l_attach_port);
	l_attach_enabled = false;
	l_attach_port = 0;
	l_attach_hostname = CIF_NULL;
    }

    /*
     * Disable event notification
     */
    l_event_listener = CIF_NULL;
    
    /*
     * Destroy the internal endpoint
     */
    globus_nexus_endpoint_destroy(&l_nx_ep);

    /*
     * Drain the queue
     */
    while(l_q_head != CIF_NULL) 
    {
	Queue_Entry *				entry;
	globus_nexus_buffer_t			buffer;

	entry = l_q_head;
	buffer = entry->buf;
	l_q_head = entry->next;
	
	globus_nexus_buffer_destroy(&buffer);
	delete entry;
    }
    l_q_tail = CIF_NULL;

    /*
     * Set the new object state
     */
    l_object_state = OBJECT_STATE_INACTIVE;

    return true;
}
/* CIF_I_Comm_Endpoint::l_deactivate() */


/*
 * CIF_I_Comm_Endpoint::deactivate(void)
 */
CIF_Bool
CIF_I_Comm_Endpoint::deactivate()
{
    CIF_Toolkit_Monitor
	monitor(l_mutex, l_deactivate_cond);

    return l_deactivate(monitor);
}
/* CIF_I_Comm_Endpoint::deactivate(void) */


/*
 * CIF_I_Comm_Endpoint::~CIF_I_Comm_Endpoint()
 */
CIF_I_Comm_Endpoint::~CIF_I_Comm_Endpoint()
{
    CIF_Comm_Status				status;
    
    CIF_Toolkit_Monitor
	monitor(l_mutex, l_deactivate_cond);

    status = CIF_Comm::status;
    {
	/*
	 * Release the endpoint, if it is still active
	 */
	l_deactivate(monitor);
	l_object_state = OBJECT_STATE_DELETED;
    }
    CIF_Comm::status = status;
}
/* CIF_I_Comm_Endpoint::~CIF_I_Comm_Endpoint() */


/*
 * CIF_I_Comm_Endpoint::i_bind_Startpoint()
 */
CIF_Bool
CIF_I_Comm_Endpoint::i_bind_Startpoint(
    globus_nexus_startpoint_t *			nx_sp)
{
    CIF_Toolkit_Monitor				monitor(l_mutex);

    int						rc;
    CIF_Comm_Status				status;

    /*
     * Verify that we have an active endpoint
     */
    status = l_verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != CIF_Comm_Status::SUCCESS)
    {
	CIF_Comm::status = status;
	return false;
	
    }

    /*
     * Bind a new startpoint to the endpoint
     */
    rc = globus_nexus_startpoint_bind(nx_sp, &l_nx_ep);

    if (rc == GLOBUS_SUCCESS)
    {
	CIF_Comm::status = CIF_Comm_Status::SUCCESS;
	return true;
    }
    else
    {
	CIF_Comm::status = CIF_Comm_Status::FAILURE;
	return false;
    }
}
/* CIF_I_Comm_Endpoint::i_bind_Startpoint() */


/*
 * CIF_I_Comm_Endpoint::l_handle_Attach()
 *
 * Handle attach requests by sending the remote process a startpoint to this
 * object's endpoint
 *
 * Paramters:
 *
 *     arg - pointer to listener class
 *
 *     my_url - the URL of this listener
 *
 *     nx_sp - pointer to a startpoint (which needs to be filled in)
 *
 */
int
CIF_I_Comm_Endpoint::l_handle_Attach(
    ARG_IN void *				arg,
    ARG_IN char *				my_url,
    ARG_IN globus_nexus_startpoint_t *		nx_sp)
{
    CIF_I_Comm_Endpoint *			ep;

    ep = (CIF_I_Comm_Endpoint *) arg;

    CIF_Toolkit_Monitor				monitor(ep->l_mutex);
    
    if (ep->l_verify_Object_State(OBJECT_STATE_ACTIVE)
	!= CIF_Comm_Status::SUCCESS)
    {
	return GLOBUS_NEXUS_ERROR_ACCEPT_ATTACH_FAILED;
    }

    if (globus_nexus_startpoint_bind(nx_sp, &ep->l_nx_ep) != GLOBUS_SUCCESS)
    {
	return GLOBUS_NEXUS_ERROR_ACCEPT_ATTACH_FAILED;
    }

    return GLOBUS_SUCCESS;
}
/* CIF_I_Comm_NSML_Endpoint::l_handle_Attach() */


/*
 * CIF_I_Comm_Endpoint::l_handle_Message() 
 *
 * receive an incoming message and queue it
 *
 * Paramters:
 *
 *     nx_ep - endpoint associated with the connection
 *
 *     nx_buf - buffer containing the message
 *
 *     hid - handler id
 *
 * Message Data:
 *
 *     char(*) - message data (completely defined by user)
 */
void
CIF_I_Comm_Endpoint::l_handle_Message(
    globus_nexus_endpoint_t *		nx_ep,
    globus_nexus_buffer_t *		nx_buf,
    int					hid)
{
    CIF_I_Comm_Endpoint *		ep;
    CIF_Toolkit_Event_Listener *	event_listener;

    ep = (CIF_I_Comm_Endpoint *)
	globus_nexus_endpoint_get_user_pointer(nx_ep);

    /*
     * Tell Nexus not to destroy this buffer
     */
    globus_nexus_buffer_save(nx_buf);
    
    /*
     * Add the buffer to the message queue and wakeup one of the waiting
     * receivers
     */
    {
	CIF_Toolkit_Monitor
	    monitor(ep->l_mutex, ep->l_q_cond);

	/*
	 * TODO: This code should be fixed once Nexus dispatch is fixed.
	 * We are making the assumption that globus_nexus_buffer_t will fit in
	 * a (void *), which it currently will, but may not in the future.
	 *
	 * Unfortunately, we had to do this because the globus_nexus_buffer_t
	 * passed to use is allocated on the stack.
	 */
	Queue_Entry *				entry;

	entry = new Queue_Entry();

	entry->buf = *nx_buf;
	entry->event_dispatched = false;
	entry->next = CIF_NULL;
	if (ep->l_q_tail == CIF_NULL)
	{
	    ep->l_q_head = entry;
	}
	else
	{
	    ep->l_q_tail->next = entry;
	}
	ep->l_q_tail = entry;

	/*
	 * If the application has an event handler registered with the
	 * endpoint, then obtain a reference to the event listener and bump the
	 * number of outstanding operations.
	 *
	 * NOTE: since we are using non-threaded handlers, it shouldn't be
	 * possible for multiple event dispatches to occur simultaneously,
	 * hence a dispatch count and thread specific state are not required.
	 */
	event_listener = ep->l_event_listener;
	if (event_listener != CIF_NULL)
	{
	    if (ep->l_event_state == EVENT_STATE_ENABLED)
	    {
		ep->l_blocking_cnt++;

		ep->l_dispatching_event = true;
		ep->l_dispatching_thread = globus_thread_self();

		entry->event_dispatched = true;
	    }
	    else
	    {
		event_listener = CIF_NULL;
	    }
	}

	monitor.signal();
    }

    /*
     * If an event listener is registered then dispatch a message arrival event
     */
    if (event_listener != CIF_NULL)
    {
	Message_Arrival_Event			event(ep);
	    
	event_listener->handle_Event(* (CIF_Toolkit_Event *) &event);
	
	{
	    CIF_Toolkit_Monitor			monitor(ep->l_mutex);

	    /*
	     * If the event listener has been disabled while we were
	     * dispatching an event, we must wake up the thread(s) that called
	     * disable_Event_Listener().
	     */
	    ep->l_dispatching_event = false;
	    
	    if (ep->l_event_state != EVENT_STATE_ENABLED)
	    {
		ep->l_event_cond.broadcast();
	    }

	    /*
	     * Decrement the number of outstanding operations, and signal the
	     * deactivation code if it is waiting to deactivate the object
	     */
	    ep->l_blocking_cnt--;

	    if (ep->l_deactivate_pending && ep->l_blocking_cnt == 0)
	    {
		ep->l_deactivate_cond.broadcast();
	    }
	}
    }
}
/* CIF_I_Comm_Endpoint::l_handle_Message() */


/*
 * CIF_I_Comm_Endpoint::l_handle_Unknown()
 *
 * handle unknown message
 *
 * Paramters:
 *
 *     ep - endpoint associated with the connection
 *
 *     buffer - buffer containing the message
 *
 *     hid - bad handler id
 *
 * Message Data:
 *
 *     none
 */
void
CIF_I_Comm_Endpoint::l_handle_Unknown(
    ARG_IN globus_nexus_endpoint_t *		nx_ep,
    ARG_INMOD globus_nexus_buffer_t *		nx_buf,
    ARG_IN int					hid)
{
    /*
     * If we are debugging, then this is definitely an error we want to catch
     */
    globus_assert(GLOBUS_FALSE);

    /*
     * However, if we are running production code, the error is ignored and
     * we'll hope that things don't completely fall over...
     */
}
/* CIF_I_Comm_Endpoint::l_handle_Unknown() */


/*
 * CIF_I_Comm_Endpoint::l_verify_Object_State()
 */
CIF_Comm_Status
CIF_I_Comm_Endpoint::l_verify_Object_State(
    Object_State			desired_state)
{
    if (l_object_state == desired_state)
    {
	return CIF_Comm_Status::SUCCESS;
    }
    
    /*
     * We aren't in the state we explected to be, so return an appropriate
     * error to the user
     */
    switch (l_object_state)
    {
      case OBJECT_STATE_DELETED:
	globus_assert(l_object_state != OBJECT_STATE_DELETED);
	return CIF_Comm_Status::OBJECT_DELETED;
	    
      case OBJECT_STATE_ACTIVE:
	return CIF_Comm_Status::OBJECT_ACTIVE;
	
      case OBJECT_STATE_INACTIVE:
	return CIF_Comm_Status::OBJECT_INACTIVE;

      default:
	return CIF_Comm_Status::INTERNAL_ERROR;
    }
}
/* CIF_I_Comm_Endpoint::l_verify_Object_State() */


/******************************************************************************
		    CIF_I_Comm_Endpoint::Properties methods
******************************************************************************/

CIF_I_Comm_Endpoint::Properties::Properties()
{
    l_proto_type = GLOBUS_NEXUS_PROTO_TYPE_ALL;
    l_proto_info_ptr = GLOBUS_NULL;
    l_proto_info_size = 0;
    l_attach_enabled = false;
    l_attach_port = 0;
}


void
CIF_I_Comm_Endpoint::Properties::set_Protocol(
    globus_nexus_proto_type_t			proto_type)
{
    l_proto_type = proto_type;
}


void
CIF_I_Comm_Endpoint::Properties::set_Protocol(
    globus_nexus_proto_type_t			proto_type,
    void *					proto_info_ptr,
    int						proto_info_size)
{
    l_proto_type = proto_type;
    l_proto_info_ptr = proto_info_ptr;
    l_proto_info_size = proto_info_size;
}


void
CIF_I_Comm_Endpoint::Properties::get_Protocol(
    globus_nexus_proto_type_t &			proto_type,
    void * &					proto_info_ptr,
    int &					proto_info_size) const
{
    proto_type = l_proto_type;
    proto_info_ptr = l_proto_info_ptr;
    proto_info_size = l_proto_info_size;
}


void
CIF_I_Comm_Endpoint::Properties::enable_Attach()
{
    l_attach_enabled = true;
}


CIF_Bool
CIF_I_Comm_Endpoint::Properties::is_Attach_Enabled() const
{
    return l_attach_enabled;
}


void
CIF_I_Comm_Endpoint::Properties::set_Attach_Port(
    const unsigned short			port)
{
    l_attach_port = port;
}


unsigned short
CIF_I_Comm_Endpoint::Properties::get_Attach_Port() const
{
    return l_attach_port;
}


/******************************************************************************
	      CIF_I_Comm_Endpoint::Message_Received_Event methods
******************************************************************************/

CIF_I_Comm_Endpoint::Message_Arrival_Event::Message_Arrival_Event(
    CIF_I_Comm_Endpoint *			ep)
{
    l_ep = ep;
}
	    
CIF_I_Comm_Endpoint *
CIF_I_Comm_Endpoint::Message_Arrival_Event::get_Endpoint() const
{
    return l_ep;
}


