/******************************************************************************
CIF_I_Comm_Listener.C

Description:

  Implementation of the CIF_Comm Listener class.  This class is used by the
  unicast protocol to establish the server side of the connection.

CVS Information:

  $Source: /home/dsl/Master/cif/CIF-Current/cpp/libraries/comm/CIF_I_Comm_Listener.C,v $
  $Date: 1999/11/15 22:51:32 $
  $Revision: 1.8 $
  $State: Exp $
  $Author: bresnaha $
******************************************************************************/

/******************************************************************************
			     Include header files
******************************************************************************/
#include "CIF_Comm.H"
#include "CIF_I_Comm_Listener.H"
#include <string.h>


/******************************************************************************
		       Define CIF_I_Comm_Listener methods
******************************************************************************/

/*
 * CIF_I_Comm_Listener::CIF_I_Comm_Listener()
 */
CIF_I_Comm_Listener::CIF_I_Comm_Listener()
{
    l_url = CIF_NULL;
    l_event_listener = CIF_NULL;
    l_blocking_cnt = 0;
    l_close_pending = false;

    l_object_state = OBJECT_STATE_INACTIVE;

    CIF_Comm::status = CIF_Comm_Status::SUCCESS;
}
/* CIF_I_Comm_Listener::CIF_I_Comm_Listener() */


CIF_Bool
CIF_I_Comm_Listener::open(
    ARG_IN const unsigned short			port,
    ARG_IN const Properties &			properties)
{
    CIF_Toolkit_Monitor				monitor(l_mutex);

    /*
     * Verify that this is an inactive but valid listener object
     */
    CIF_Comm_Status				status;

    status = l_verify_Object_State(OBJECT_STATE_INACTIVE);
    if (status != CIF_Comm_Status::SUCCESS)
    {
	CIF_Comm::status = status;
	return false;
    }

    /*
     * Process properties before establishing an active endpoint
     */
    l_event_listener = properties.get_Event_Listener();
    
    /*
     * Create an endpoint which will be used to receive messages from the
     * process attempting to connect to this listener
     */
    CIF_I_Comm_Endpoint::Properties		ep_properties;

    if (l_event_listener != CIF_NULL)
    {
	ep_properties.set_Event_Listener((CIF_Toolkit_Event_Listener *) this);
    }
    ep_properties.enable_Attach();
    ep_properties.set_Attach_Port(port);
    
    if (l_ep.activate(ep_properties) == false)
    {
/*	if (CIF_Comm::status != CIF_Comm_Status::FAILURE)
	{
	    CIF_Comm::status = CIF_Comm_Status::INTERNAL_ERROR;
	}
*/	return false;
    }
    
    /*
     * Create a URL string that the application can use
     */
    l_url = new char [(strlen(get_URL_Scheme()) +
		       strlen(l_ep.get_Attach_Hostname()) + 14)];
    sprintf(l_url,
	    "%s://%s:%d",
	    get_URL_Scheme(),
	    l_ep.get_Attach_Hostname(),
	    l_ep.get_Attach_Port());

    /*
     * Update object state and return
     */
    l_object_state = OBJECT_STATE_ACTIVE;

    CIF_Comm::status = CIF_Comm_Status::SUCCESS;
    return true;

  fn_abort:
    l_event_listener = CIF_NULL;
    CIF_Comm::status = CIF_Comm_Status::FAILURE;
    return false;
}
/* CIF_I_Comm_Listener::open() */


/*
 * CIF_I_Comm_Listener::get_Connection()
 */
CIF_Comm_Connection *
CIF_I_Comm_Listener::get_Connection()
{
    return l_get_Connection(CIF_Comm_Connection::Properties(),
			    CIF_Toolkit_Timeout::INFINITY);
}


/*
 * CIF_I_Comm_Listener::get_Connection(timeout)
 */
CIF_Comm_Connection *
CIF_I_Comm_Listener::get_Connection(
    ARG_IN const CIF_Toolkit_Timeout &		timeout)
{
    return l_get_Connection(CIF_Comm_Connection::Properties(),
			    timeout);
}


/*
 * CIF_I_Comm_Listener::get_Connection(properties)
 */
CIF_Comm_Connection *
CIF_I_Comm_Listener::get_Connection(
    ARG_IN const
    CIF_Comm_Connection::Properties &		properties)
{
    return l_get_Connection(properties,
			    CIF_Toolkit_Timeout::INFINITY);
}
    

/*
 * CIF_I_Comm_Listener::get_Connection(properties, timeout)
 */
CIF_Comm_Connection *
CIF_I_Comm_Listener::get_Connection(
    ARG_IN const
    CIF_Comm_Connection::Properties &		properties,
    ARG_IN const CIF_Toolkit_Timeout &		timeout)
{
    return l_get_Connection(properties,
			    timeout);
}
    

/*
 * CIF_I_Comm_Listener::l_get_Connection()
 */
CIF_Comm_Connection *
CIF_I_Comm_Listener::l_get_Connection(
    ARG_IN const
    CIF_Comm_Connection::Properties &		properties,
    ARG_IN const CIF_Toolkit_Timeout &		timeout)
{
    CIF_Bool					rc;
    CIF_Comm_Connection *			conn;
    nexus_buffer_t				buf;
    
    {
	CIF_Toolkit_Monitor			monitor(l_mutex);

	/*
	 * Verify that this is an inactive but valid listener object
	 */
	CIF_Comm_Status				status;

	status = l_verify_Object_State(OBJECT_STATE_ACTIVE);
	if (status != CIF_Comm_Status::SUCCESS)
	{
	    CIF_Comm::status = status;
	    return CIF_NULL;
	}

	l_blocking_cnt++;
    }

    rc = l_ep.receive(buf, timeout);
    
    {
	CIF_Toolkit_Monitor			monitor(l_mutex);

	l_blocking_cnt--;
	if (l_close_pending && l_blocking_cnt == 0)
	{
	    l_close_cond.signal();
	}
    }
    
    if (rc == false)
    {
	return CIF_NULL;
    }

    /*
     * Establish connection with remote end
     */
    conn = open_Connection(buf, properties, timeout);
    globus_nexus_buffer_destroy(&buf);

    return conn;
}
/* CIF_I_Comm_Listener::l_get_Connection() */


/*
 * CIF_I_Comm_Listener::enable_Event_Listener()
 */
CIF_Bool
CIF_I_Comm_Listener::enable_Event_Listener()
{
    CIF_Bool					rc;
    
    {
	CIF_Toolkit_Monitor			monitor(l_mutex);

	/*
	 * Verify that this is an inactive but valid listener object
	 */
	CIF_Comm_Status				status;

	status = l_verify_Object_State(OBJECT_STATE_ACTIVE);
	if (status != CIF_Comm_Status::SUCCESS)
	{
	    CIF_Comm::status = status;
	    return false;
	}

	l_blocking_cnt++;
    }
    
    rc = l_ep.enable_Event_Listener();

    {
	CIF_Toolkit_Monitor			monitor(l_mutex);

	l_blocking_cnt--;
	if (l_close_pending && l_blocking_cnt == 0)
	{
	    l_close_cond.signal();
	}
    }
    
    return rc;
}


/*
 * CIF_I_Comm_Listener::disable_Event_Listener()
 */
CIF_Bool
CIF_I_Comm_Listener::disable_Event_Listener()
{
    CIF_Bool					rc;
    
    {
	CIF_Toolkit_Monitor			monitor(l_mutex);

	/*
	 * Verify that this is an inactive but valid listener object
	 */
	CIF_Comm_Status				status;

	status = l_verify_Object_State(OBJECT_STATE_ACTIVE);
	if (status != CIF_Comm_Status::SUCCESS)
	{
	    CIF_Comm::status = status;
	    return false;
	}

	l_blocking_cnt++;
    }
    
    rc = l_ep.disable_Event_Listener();
    
    {
	CIF_Toolkit_Monitor			monitor(l_mutex);

	l_blocking_cnt--;
	if (l_close_pending && l_blocking_cnt == 0)
	{
	    l_close_cond.signal();
	}
    }
    
    return rc;
}


/*
 * CIF_I_Comm_Listener::get_URL()
 */
const char *
CIF_I_Comm_Listener::get_URL()
{
    CIF_Toolkit_Monitor				monitor(l_mutex);

    /*
     * Verify that this is a valid listener object
     */
    CIF_Comm_Status				status;

    status = l_verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != CIF_Comm_Status::SUCCESS)
    {
	CIF_Comm::status = status;
	return CIF_NULL;
    }

    return l_url;
}


/*
 * CIF_I_Comm_Listener::l_close()
 */
CIF_Bool
CIF_I_Comm_Listener::l_close(
    CIF_Toolkit_Monitor &			monitor)
{

    CIF_Comm_Status				status;

    /*
     * If the listener is current active, then deactivate it; otherwise return
     * an error
     */
    status = l_verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != CIF_Comm_Status::SUCCESS)
    {
	CIF_Comm::status = status;
	return false;
    }

    /*
     * If this is the first thread to attempt a close, then close down the
     * endpoint so that blocking operations will return
     */
    if (l_close_pending == false)
    {
	l_close_pending = true;
	l_ep.deactivate();
    }

    /*
     * Wait for all outstanding blocking operations to complete
     */
    while (l_blocking_cnt > 0)
    {
	monitor.wait();

	/*
	 * Verify that another thread hasn't closed the listener while we were
	 * sleeping
	 */
	status = l_verify_Object_State(OBJECT_STATE_ACTIVE);
	if (status != CIF_Comm_Status::SUCCESS)
	{
	    CIF_Comm::status = status;
	    return false;
	}
    }
    
    /*
     * Reset state to indicate that the connection is closed
     */
    if (l_url != CIF_NULL)
    {
	delete [] l_url;
	l_url = CIF_NULL;
    }

    l_close_pending = false;
    l_event_listener = CIF_NULL;
    
    l_object_state = OBJECT_STATE_INACTIVE;

    /*
     * Wake up any remaining closers so that they can error out
     */
    l_close_cond.broadcast();

    return true;
}
/* CIF_I_Comm_Listener::l_close() */

/*
 * CIF_I_Comm_Listener::close()
 */
CIF_Bool
CIF_I_Comm_Listener::close()
{
    CIF_Toolkit_Monitor
	monitor(l_mutex, l_close_cond);

    return l_close(monitor);
}
/* CIF_I_Comm_Listener::close() */


/*
 * CIF_I_Comm_Listener::~CIF_I_Comm_Listener()
 */
CIF_I_Comm_Listener::~CIF_I_Comm_Listener()
{
    CIF_Comm_Status				status;
    
    CIF_Toolkit_Monitor
	monitor(l_mutex, l_close_cond);

    status = CIF_Comm::status;
    {
	l_close(monitor);

	l_object_state = OBJECT_STATE_DELETED;
    }
    CIF_Comm::status = status;
}
/* CIF_I_Comm_Listener::~CIF_I_Comm_Listener() */


/*
 * CIF_I_Comm_Listener::l_verify_Object_State()
 */
CIF_Comm_Status
CIF_I_Comm_Listener::l_verify_Object_State(
    ARG_IN const Object_State &			desired_state)
{
    if (l_object_state == desired_state)
    {
	return CIF_Comm_Status::SUCCESS;
    }
    
    /*
     * We aren't in the state we explected to be, so return an appropriate
     * error to the user
     */
    switch (l_object_state)
    {
      case OBJECT_STATE_DELETED:
	globus_assert(l_object_state != OBJECT_STATE_DELETED);
	return CIF_Comm_Status::OBJECT_DELETED;
	    
      case OBJECT_STATE_ACTIVE:
	return CIF_Comm_Status::OBJECT_ACTIVE;
	
      case OBJECT_STATE_INACTIVE:
	return CIF_Comm_Status::OBJECT_INACTIVE;

      default:
	return CIF_Comm_Status::INTERNAL_ERROR;
    }
}
/* CIF_I_Comm_Listener::l_verify_Object_State() */


/*
 * CIF_I_Comm_Listener::handle_Event()
 */
void
CIF_I_Comm_Listener::handle_Event(
    ARG_INMOD CIF_Toolkit_Event &		ep_event)
{
    {
	CIF_Toolkit_Monitor			monitor(l_mutex);

	/*
	 * Verify that this is an active connection
	 */
	CIF_Comm_Status				status;

	status = l_verify_Object_State(OBJECT_STATE_ACTIVE);
	globus_assert(status == CIF_Comm_Status::SUCCESS);
	if (status != CIF_Comm_Status::SUCCESS)
	{
	    return;
	}

	l_blocking_cnt++;
    }

    Connection_Request_Event			cr_event(this);
    
    l_event_listener->handle_Event(cr_event);

    {
	CIF_Toolkit_Monitor			monitor(l_mutex);

	l_blocking_cnt--;
	if (l_close_pending && l_blocking_cnt == 0)
	{
	    l_close_cond.signal();
	}
    }
}
	
