/******************************************************************************
Listener.java

CVS Information:

  $Source: /home/dsl/Master/cif/CIF-Current/java/libraries/cif/comm/Listener.java,v $
  $Date: 1998/10/13 06:33:47 $
  $Revision: 1.3 $
  $State: Exp $
  $Author: toonen $
******************************************************************************/

package cif.comm;

import cif.toolkit.*;

/**
 * An abstract interface for a CIF Listener object.  New listener objects
 * are created using the CIF Factory class.
 *
 * @see cif.comm.Factory
 */
public
interface Listener
{
    /**
     * Enable the listener to deliver events to the event listener.  Note,
     * this is necessary to avoid race conditions associated with events
     * arriving before the application is capable of dealing with them.  If
     * events occur before this method is called, the events are queued and
     * delivered once event notification is enabled.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the connection object has not
     *     been previously activated or has been closed (deactivated).
     */
    public
    void
    enable_Event_Notification()
    throws
	Object_Inactive_Exception;
	
    /**
     * Stop delivering events to the event listener.  Note, this method will
     * block until all in-progress event dispatching associated with this
     * listener has completed.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the connection object has not
     *     been previously activated or has been closed (deactivated).
     */
    public
    void
    disable_Event_Notification()
    throws
	Object_Inactive_Exception;
	
    /**
     * Obtain the next established connection.  If no connection is
     * established, this function blocks until such time as one is established.
     *
     * @see cif.comm.Listener#get_Connection(Connection_Properties, cif.toolkit.Timeout)
     */
    public
    Connection
    get_Connection()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException;
    
    /**
     * Obtain the next established connection.  If no connection is
     * established, this function blocks until such time as one is established
     * or the specified amount of time elapses.
     *
     * @param timeout
     *
     *     If a new connection is not immediately available, wait up to timeout
     *     number of millseconds for it to arrive.
     *
     * @see cif.comm.Listener#get_Connection(Connection_Properties, cif.toolkit.Timeout)
     */
    public
    Connection
    get_Connection(
	long					timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException;
    
    /**
     * Obtain the next established connection.  If no connection is
     * established, this function blocks until such time as one is established
     * or the specified amount of time elapses.
     *
     * @see cif.comm.Listener#get_Connection(Connection_Properties, cif.toolkit.Timeout)
     */
    public
    Connection
    get_Connection(
	cif.toolkit.Timeout			timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException;
    

    /**
     * Obtain the next established connection.  If no connection is
     * established, this function blocks until such time as one is established.
     *
     * @see cif.comm.Listener#get_Connection(Connection_Properties, cif.toolkit.Timeout)
     */
    public
    Connection
    get_Connection(
	Connection_Properties			properties)
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException;
    
    /**
     * Obtain the next established connection.  If no connection is
     * established, this function blocks until such time as one is established
     * or the specified amount of time elapses.
     *
     * @see cif.comm.Listener#get_Connection(Connection_Properties, cif.toolkit.Timeout)
     */
    public
    Connection
    get_Connection(
	Connection_Properties			properties,
	long					timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException;
    
    /**
     * Obtain the next established connection.  If no connection is
     * established, this function blocks until such time as one is established
     * or the specified amount of time elapses.
     *
     * @param properties
     *
     *    Properties to be applied to the newly formed connection
     *
     * @param timeout
     *
     *     If a new connection is not immediately available, wait up to the
     *     point where the timeout object expires for it to arrive.
     *
     * @return
     *
     *     A connection object for the newly formed connection.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the connection object has not
     *     been previously activated or has been closed (deactivated).
     *
     * @exception cif.toolkit.Timeout_Expired_Exception
     *
     *     Timeout_Expired_Exception is thrown if the specified amount of time
     *     elapses before a message arrives.
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     *
     * @exception cif.comm.Comm_Failure_Exception
     *
     *     Comm_Failure_Exception is thrown if the underlying connection was
     *     terminated due to some communication breakdown.  This implies that
     *     one or more of the messages sent by this object may not have been
     *     delivered to the other objects participating in the connection, and
     *     that some of the messages sent by those objects may not have been
     *     received by this object.  However, the exact implications of the
     *     failure are dependent on the underlying protocol.
     *
     * @exception cif.comm.Comm_Would_Block_Exception
     *
     *     Comm_Would_Block_Exception is thrown if the receive() method was
     *     called by the same thread currently handling a CIF event, and the
     *     receive() method would have blocked waiting for a message to arrive.
     *     If called from an event handler, receive() may only be called with a
     *     timeout value of zero.
     *
     * @exception InterruptedException
     *
     *     InterruptedException is thrown if the thread was interrupted for
     *     some reason.  Presumably, the user's code intentionally performed
     *     the interrupt, so the exception is passed through.
     */
    public
    Connection
    get_Connection(
	Connection_Properties			properties,
	cif.toolkit.Timeout			timeout)
    throws
	Object_Inactive_Exception,
	Timeout_Expired_Exception,
	Comm_Inactive_Exception,
	Comm_Failure_Exception,
	Comm_Would_Block_Exception,
	InterruptedException;
    

    /**
     * Obtain the URL which can be used to form a new connection.
     *
     * @return
     *
     *     The URL which can be used to form a new connection with the
     *     listener.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the connection object has not
     *     been previously activated or has been closed (deactivated).
     */
    public
    String
    get_URL()
    throws
	Object_Inactive_Exception;

    
    /**
     * Close (deactivate) the listener.
     *
     * @exception cif.toolkit.Object_Inactive_Exception
     *
     *     Object_Inactive_Exception is thrown if the connection object has not
     *     been previously activated or has been closed (deactivated).
     *
     * @exception cif.comm.Comm_Inactive_Exception
     *
     *     Comm_Inactive_Exception is thrown if Comm has not been properly
     *     activated
     *
     */
    public
    void
    close()
    throws
	Object_Inactive_Exception,
	Comm_Inactive_Exception;
}
