package cif.shared_state;

import java.lang.*;
import nexus.*;


//**************************************************************************
//*  L_Buffer_Holder
//*  ---------------
//**************************************************************************
class L_Buffer_Holder
{
  public int  offset;
  public byte buf[];
}

//*************************************************************************
//*  L_Buffer_List_Node
//*  ------------------
//*************************************************************************
class L_Buffer_List_Node
{
  protected int                  local_Size;
  protected int                  current_Offset;
  protected byte                 local_Buffer[];

  public L_Buffer_List_Node  next;

  /*
   *
   */
  public L_Buffer_List_Node(int size)
  {
    local_Buffer = new byte[size];
    local_Size = size;
    current_Offset = 0;
  }

  /*
   *
   */
  public L_Buffer_List_Node(byte value[], 
                            int size)
  {
    local_Buffer = new byte[size];
    System.arraycopy(value, 0, local_Buffer, 0, size);
    local_Size = size;
    current_Offset = 0;
  }

  /*
   *
   */
  public synchronized int  
  copy_Buffer(
      byte                         buf[], 
      int                          offset)
  {
    System.arraycopy(local_Buffer, 0, buf, offset, current_Offset);
    return local_Size;
  }

  /*
   *
   */
  public synchronized byte [] 
  get_Buffer()
  {
     return local_Buffer;
  }

  /*
   *
   */
  public synchronized int 
  get_User_Buffer(
      int                        size_requested)
  {
    int old_offset;

    if(size_requested + current_Offset > local_Size)
    {
      local_Size = current_Offset;
      return -1;
    }
    old_offset = current_Offset;
    current_Offset += size_requested;

    return old_offset;
  }
};

//*************************************************************************
//*  L_Buffer_Manager
//*  ----------------
//*************************************************************************
class L_Buffer_Manager
{
  protected int            size_Used;
  protected int            current_Size;
  protected int            node_Size;

  protected L_Buffer_List_Node   first_Node;
  protected L_Buffer_List_Node   current_Node;

  /*
   *
   */
  public L_Buffer_Manager(int size)
  {
    first_Node = new L_Buffer_List_Node(size);
    current_Node = first_Node;

    size_Used = 0;
    current_Size = size;
    node_Size = size;
  }

  /*
   *
   */
  public L_Buffer_Manager(byte in_value[], 
                          int size)
  {
    first_Node = new L_Buffer_List_Node(in_value, size);
    current_Node = first_Node;

    size_Used = size-1;
    current_Size = size;
    node_Size = size;
  }

  /*
   *
   */
  public synchronized int 
  get_Size() 
  {
    return size_Used;
  }
             
  /*
   *
   */
  public synchronized byte [] 
  get_Buffer()
  {
    byte buf[];
    int  offset = 0;
    int  len;
    L_Buffer_List_Node i;

    if(current_Node == first_Node)
    {
      return current_Node.get_Buffer();
    }

    buf = new byte[size_Used];
    for(i = first_Node; i != null; i = i.next)
    {
      len = i.copy_Buffer(buf, offset);
      offset += len;
    }
    return buf;
  }

  /*
   *
   */
  public synchronized L_Buffer_Holder 
  get_Pack_Buffer(int size_requested)
  {
    int  offset;
    L_Buffer_Holder holder = new L_Buffer_Holder();

    offset = current_Node.get_User_Buffer(size_requested);
    node_Size += size_requested;

    while(offset < 0)
    {
      current_Node.next = new L_Buffer_List_Node(node_Size);
      current_Size += node_Size;
      current_Node = current_Node.next;
    
      offset = current_Node.get_User_Buffer(size_requested);
    }
   
    holder.offset = offset;
    holder.buf = current_Node.get_Buffer();
    size_Used += size_requested;

    return holder;
  }

  /*
   *
   */
  public L_Buffer_Holder get_Unpack_Buffer(int size_requested)
                         throws BufferOverrunException
  {
    int  offset;
    L_Buffer_Holder holder = new L_Buffer_Holder();

    offset = current_Node.get_User_Buffer(size_requested);
    node_Size += size_requested;

    if(offset < 0)
    {
	throw new BufferOverrunException();
    }
   
    holder.offset = offset;
    holder.buf = current_Node.get_Buffer();
    size_Used += size_requested;

    return holder;
  }

};


//*********************************************************************
//*  CIF.Shared_State.Serial
//*  -----------------------
//**********************************************************************

/**
   The serial class provides convienience methods for packing 
   primative data types into machine independent byte arrays.
    
<pre>  
   The Serial object handles the needed data conversions between different
   platforms.  Primative data types can be packed on a machine using Java byte
   formating.  The user can then call "get_Buffer()" which returns a 
   byte array with information about the machine on wich it was packed.
   This byte array can the be sent accross a wire to a program written
   with the CIF_Shared_State C++ libraries and using Big Endian formatting.
   This byte array can then be used to instatiate a new Serial object within
   the C++ program and user can use the unpack methods to obtatin the
   primative data types packed in by the Java program.
</pre> 
*/
public class Serial
{
    private L_Buffer_Manager buffer_Manager;

    private static int L_Serial_Message_Size = 1024;
    private int local_Format;
    
    /**
     Creates a new empty serial object.
     */
    public Serial()
    {
        L_Buffer_Holder holder;

        buffer_Manager = new L_Buffer_Manager(L_Serial_Message_Size);
        holder = buffer_Manager.get_Pack_Buffer(1);
        local_Format = DataConversion.NEXUS_DC_FORMAT_LOCAL;
        holder.buf[holder.offset] = (byte) local_Format;

    }

    /**
       Creates a new Serial object and sets its value the the passed
       byte array.  Within the byte array is a header format. 
       This information is used by the Serial object for data conversion.
     */
    public Serial(byte value[], int size)
    { 
        L_Buffer_Holder holder;

        buffer_Manager = new L_Buffer_Manager(value, size);
        holder = buffer_Manager.get_Pack_Buffer(1);
        local_Format = (int)holder.buf[holder.offset];        
    }

    /**
       Returns the machine independent byte array of the Serial object.
       The first byte is use to determin the data format, the remaning
       bytes contain user.
     */
    public synchronized byte [] 
    get_Buffer()
    {
	return buffer_Manager.get_Buffer();
    }
    
    /**
       Returns the size of the buffer.  This is not always equal to
       the length of the buffer array.
     */
    public synchronized int 
    get_Size_Buffer()
    {
	int size;
	
	size = buffer_Manager.get_Size();
	if(L_Serial_Message_Size < size)
	{
	    L_Serial_Message_Size = size;
	}
	return  size;
    }

    /**
       Retrieve a serialized array of bytes.
       @param value byte array where data will be unpacked
       @param offset index into value where unpacking should begin.
       @param count  the number of bytes to unpack
     */
    public void 
    unpack_Byte(
        byte                                value[], 
        int                                 offset, 
        int                                 count)
     throws LostPrecisionException,
	    BufferOverrunException
   {
       L_Buffer_Holder holder;
       int  size_requested;
       
       size_requested = DataConversion.sizeof_byte(count, local_Format);
       holder = buffer_Manager.get_Unpack_Buffer(size_requested);
       
       DataConversion.get_byte(value, offset, count, 
			       holder.buf, holder.offset + offset,
			       local_Format);
   }

    /**
       Retrieve a serialized array of chars.
       @param value char array where data will be unpacked
       @param offset index into byte where unpacking should begin.
       @param count  the number of chars to unpack
     */
    public void unpack_Char(char value[], 
			    int offset, 
			    int count)
	throws LostPrecisionException,
	       BufferOverrunException
	       
    {
	L_Buffer_Holder holder;
	int  size_requested;
	
	size_requested = DataConversion.sizeof_char(count, local_Format);
	holder = buffer_Manager.get_Unpack_Buffer(size_requested);
	
	DataConversion.get_char(value, offset, count, 
				holder.buf, holder.offset, 
				local_Format);
	
    }

    /**
       Retrieve a serialized array of booleans.
       @param value boolean array where data will be unpacked
       @param offset index into value where unpacking should begin.
       @param count  the number of booleans to unpack
     */
    public void unpack_Bool(boolean value[],
			    int offset, 
			    int count)
	throws LostPrecisionException,
	       BufferOverrunException
    {
	L_Buffer_Holder holder;
	int  bytes_needed;
	byte byte_array[];
	int  ctr;
	int  byte_array_size;
	int  size_requested;
	int  index;
	int  decoded;
	byte mask;
	
	byte_array_size = count / (8) + 1;
	byte_array = new byte[byte_array_size];
	
	size_requested = DataConversion.sizeof_byte(byte_array_size, local_Format);
	holder = buffer_Manager.get_Unpack_Buffer(size_requested);
	
	DataConversion.get_byte(byte_array, 0, byte_array_size, 
				holder.buf, holder.offset, 
				local_Format);
	
	index = 0;
	decoded = 0;
	while(decoded < count)
	{
	    mask = 1;
	    for(ctr = 0; ctr < 8 && decoded < count; ctr++)
	    {
		if((byte_array[index] & mask) != 0)
		{
		    value[offset + decoded] = true;
		}
		else
		{
		    value[offset + decoded] = false;
		}
		decoded++;
		mask =(byte) (mask << 1);
	    }
	    index++;
	}
    }

      /**
       Retrieve a serialized array of shorts.
       @param value short array where data will be unpacked
       @param offset index into value where unpacking should begin.
       @param count  the number of shorts to unpack
      */
    public void unpack_Short(short value[],
			     int offset, 
			     int count)
	throws LostPrecisionException,
	       BufferOverrunException
    {
	L_Buffer_Holder holder;
	int  size_requested;
	
	size_requested = DataConversion.sizeof_short(count, local_Format);
	holder = buffer_Manager.get_Unpack_Buffer(size_requested);
	
	DataConversion.get_short(value, offset, count, 
				 holder.buf, holder.offset, 
				 local_Format);
    }

    /**
       Retrieve a serialized array of ints.
       @param value int array where data will be unpacked
       @param offset index into value where unpacking should begin.
       @param count  the number of ints to unpack
    */
    public void unpack_Int(int value[],
                           int offset, 
                           int count)
	throws LostPrecisionException,
	       BufferOverrunException
    {
	L_Buffer_Holder holder;
	int  size_requested;
	
	size_requested = DataConversion.sizeof_int(count, local_Format);
	holder = buffer_Manager.get_Unpack_Buffer(size_requested);
	
	DataConversion.get_int(value, offset, count, 
			       holder.buf, holder.offset, 
			       local_Format);
   }
    
    /**
       Retrieve a serialized array of longs.
       @param value long array where data will be unpacked
       @param offset index into value where unpacking should begin.
       @param count  the number of longs to unpack
    */
    public void unpack_Long(long value[],
                           int offset, 
                           int count)
                           throws LostPrecisionException,
				  BufferOverrunException
   {
     L_Buffer_Holder holder;
     int  size_requested;

     size_requested = DataConversion.sizeof_long(count, local_Format);
     holder = buffer_Manager.get_Unpack_Buffer(size_requested);

     DataConversion.get_long(value, offset, count, 
                             holder.buf, holder.offset, 
                             local_Format);
   }
    
    /**
       Retrieve a serialized array of floats.
       @param value float array where data will be unpacked
       @param offset index into value where unpacking should begin.
       @param count  the number of floats to unpack
    */
    public void unpack_Float(float value[],
			     int offset, 
			     int count)
	throws LostPrecisionException,
	       BufferOverrunException
    {
	L_Buffer_Holder holder;
	int  size_requested;
	
	size_requested = DataConversion.sizeof_float(count, local_Format);
	holder = buffer_Manager.get_Unpack_Buffer(size_requested);
	
	DataConversion.get_float(value, offset, count, 
				 holder.buf, holder.offset, 
				 local_Format);
    }

    /**
       Retrieve a serialized array of doubles.
       @param value double array where data will be unpacked
       @param offset index into value where unpacking should begin.
       @param count  the number of doubles to unpack
    */
    public void unpack_Double(double value[],
			      int offset, 
			      int count)
	throws LostPrecisionException,
	       BufferOverrunException
    {
	L_Buffer_Holder holder;
	int  size_requested;
	
	size_requested = DataConversion.sizeof_double(count, local_Format);
	holder = buffer_Manager.get_Unpack_Buffer(size_requested);
	
	DataConversion.get_double(value, offset, count, 
				  holder.buf, holder.offset, 
				  local_Format);
   }

    /**
       Serialize an array of bytes.
       @param value byte array to pack into serial object
       @param offset index into value where packing should begin.
       @param count  the number of bytes to pack in
    */
    public void pack_Byte(byte value[],
			  int offset, 
			  int count)
    {
	int size_requested;
	L_Buffer_Holder holder;
	
	size_requested = DataConversion.sizeof_byte(count, local_Format);
	holder = buffer_Manager.get_Pack_Buffer(size_requested);
	
	DataConversion.put_byte(value, offset, count, 
				holder.buf, holder.offset); 
    }
    
    /**
       Serialize an array of chars.
       @param value char array to pack in
       @param offset index into value where packing should begin.
       @param count  the number of chars to pack in
    */
    public void pack_Char(char value[],
			  int offset, 
			  int count)
    {
	int size_requested;
	L_Buffer_Holder holder;
	
	size_requested = DataConversion.sizeof_char(count, local_Format);
	holder = buffer_Manager.get_Pack_Buffer(size_requested);
	
	DataConversion.put_char(value, offset, count, holder.buf, holder.offset); 
    }

    /**
       Serialize an array of booleans.
       @param value boolean array to pack in
       @param offset index into value where packing should begin.
       @param count  the number of booleans to pack in
    */
    public void pack_Bool(boolean value[],
			  int offset, 
			  int count)
    {
	int  size_requested;
	byte byte_array[];
	int  byte_array_size;
	int  encoded;
	byte mask;
	int  index;
	int  ctr;
	L_Buffer_Holder holder;
	
	
	byte_array_size = count / 8 + 1;
	byte_array = new byte[byte_array_size];
	
	index = 0;
	encoded = 0;
	while(encoded < count)
	{
	    byte_array[index] = 0;
	    mask = 1;
	    for(ctr = 0; ctr < 8 && encoded < count; ctr++)
	    {
		if(value[encoded+offset])
		{
		    byte_array[index] = (byte) (byte_array[index] | mask);
		}
		mask = (byte) (mask << 1);
		encoded++;
	    }
	    index++;
	}

	size_requested = DataConversion.sizeof_byte(byte_array_size, 
						    local_Format);
	holder = buffer_Manager.get_Pack_Buffer(size_requested);
	
	DataConversion.put_byte(byte_array, 0, byte_array_size, 
				holder.buf, holder.offset); 
    }

    /**
       Serialize an array of shorts.
       @param value boolean array to pack in
       @param offset index into value where packing should begin.
       @param count  the number of shorts to pack into the serial object.
    */
    
    public void pack_Short(short value[],
			   int offset, 
			   int count)
    {
	int size_requested;
	L_Buffer_Holder holder;
	
	size_requested = DataConversion.sizeof_short(count, local_Format);
	holder = buffer_Manager.get_Pack_Buffer(size_requested);
	
	DataConversion.put_short(value, offset, count, 
				 holder.buf, holder.offset); 
    }

    /**
       Serialize an array of ints.
       @param value int array to pack in
       @param offset index into value where packing should begin.
       @param count  the number of ints to pack into the serial object.
    */
    public void pack_Int(int value[],
			 int offset, 
			 int count)
    {
	int size_requested;
	L_Buffer_Holder holder;
	
	size_requested = DataConversion.sizeof_int(count, local_Format);
	holder = buffer_Manager.get_Pack_Buffer(size_requested);
	
	DataConversion.put_int(value, offset, count, 
			       holder.buf, holder.offset); 
    }

    /**
       Serialize an array of longs.
       @param value long array to pack into serial object.
       @param offset index into value where packing should begin.
       @param count  the number of ints to pack into the serial object.
    */
    public void pack_Long(long value[],
			  int offset, 
			  int count)
    {
	int size_requested;
	L_Buffer_Holder holder;
	
	size_requested = DataConversion.sizeof_long(count, local_Format);
	holder = buffer_Manager.get_Pack_Buffer(size_requested);
	
	DataConversion.put_long(value, offset, count, 
				holder.buf, holder.offset); 
    }

    /**
       Serialize an array of floats.
       @param value floats array to pack into serial object.
       @param offset index into value where packing should begin.
       @param count  the number of floats to pack into the serial object.
    */
    public void pack_Float(float value[],
			   int offset, 
			   int count)
    {
	int size_requested;
	L_Buffer_Holder holder;
	
	size_requested = DataConversion.sizeof_float(count, local_Format);
	holder = buffer_Manager.get_Pack_Buffer(size_requested);
	
	DataConversion.put_float(value, offset, count, 
				 holder.buf, holder.offset); 
    }

    /**
       Serialize an array of doubles.
       @param value doubles array to pack into serial object.
       @param offset index into value where packing should begin.
       @param count  the number of doubles to pack into the serial object.
    */
    public void pack_Double(double value[],
			    int offset, 
			    int count)
    {
	int size_requested;
	L_Buffer_Holder holder;
	
	size_requested = DataConversion.sizeof_double(count, local_Format);
	holder = buffer_Manager.get_Pack_Buffer(size_requested);
	
	DataConversion.put_double(value, offset, count, 
				  holder.buf, holder.offset); 
    }
};

