package cif.shared_state;

import cif.shared_state.messenger.*;
import java.lang.*;
import java.util.*;
import java.net.*;
import nexus.*;
/**
   A Set class contains and manages States across message space.
<blockquote>  

Each member of shared state message space has a local copy of a Set.
On instantiation each set connects to message space.  The user adds States to
the set via Set.add_Shared_State().  Once a State is added the Set will 
keep it synchronized with all other States of the same name in all the 
other Sets connected to the same message space.
</blockquote> 
*/

//*************************************************************************
//*  Set
//*  ---
//*************************************************************************
public class 
Set 
  implements Set_Event_Listener
{
    private String                message_Space_Url;
    private Hashtable             hash_Table;
    private Messenger_Interface   messenger;
    private Set_GUID              guid;
    private Vector                listeners;
    

    /** An indentifier telling the Set to use a dummy messaging protocol.*/    
    public static final int         CIF_MESSENGER_TYPE_DUMMY = 0;
    /** An indentifier telling the Set to use a tcp messaging protocol.*/    
    public static final int         CIF_MESSENGER_TYPE_TCP = 1;
    
    /**
       Create the first set in message space.
       @param message_type indicates what type of messenger to use
       Current options are CIF_MESSENGER_TYPE_DUMMY and 
       CIF_MESSENGER_TYPE_TCP.
     */
    public 
    Set(
        int                          messenger_type) 
      throws Exception
    {
        init();
        switch(messenger_type)
	{
            case CIF_MESSENGER_TYPE_DUMMY:
	        messenger = new Messenger_Dummy(this);
	        break;
	    
	    case CIF_MESSENGER_TYPE_TCP:
	        messenger = new Messenger_TCP(this);
	        break;
        }
	
        message_Space_Url = messenger.get_Message_Space_URL();
    }
    
    /**
    *  Create a Set and connect to message space via the passed in url.
    *  @param url 
    *    address of the message space connection listener.
    *
    *  @param
    *    message_type indicates what type of messenger to use
    *       Current options are CIF_MESSENGER_TYPE_DUMMY and 
    *       CIF_MESSENGER_TYPE_TCP.
    */
    public 
    Set( 
        int                                messenger_type, 
        String                             url) 
	throws Exception
    {
        init();
        switch(messenger_type)
	{
	case CIF_MESSENGER_TYPE_DUMMY:
	    messenger = new Messenger_Dummy(this);
	    break;
	    
	case CIF_MESSENGER_TYPE_TCP:
	    messenger = new Messenger_TCP(this, url);
	    break;
        }
	
        message_Space_Url = messenger.get_Message_Space_URL();	
    }
    
    /**
    *  Add a Set_Event_Listener to the list of objects notified
    *  when a Set joins, leaves, or changes a state in message space.
    */
    public synchronized void 
    add_Set_Event_Listener(
        Set_Event_Listener                        l)
    {
        listeners.addElement(l);
    }
    
    /**
    *  Remove a Set_Event_Listener to the list of objects notified
    *  when a Set joins, leaves, or changes a state in message space.
    */
    public synchronized void 
    remove_Set_Event_Listener(
        Set_Event_Listener                       l)
    {
        listeners.removeElement(l);
    }
    
    /*
     *  init()
     */
    private void init() 
	throws Exception
    {
        byte guid_bytes[] = new byte[0];
        byte randID[] = new byte[0];
        int  len = 0;
	
        try 
	{
            byte hostID[];
            byte time_bytes[];
            Date date = new Date();
	    
            InetAddress address;
            address = InetAddress.getLocalHost();
            time_bytes = new Long(date.getTime()).toString().getBytes();
	    
            randID = new Long(new Random().nextLong()).toString().getBytes();

            hostID = address.getAddress();
            len = hostID.length + time_bytes.length + randID.length; 
            guid_bytes = new byte[len];
            System.arraycopy(hostID, 0, guid_bytes, 0, hostID.length);
            System.arraycopy(time_bytes, 0, guid_bytes, 
                             hostID.length, time_bytes.length);
            System.arraycopy(randID, 0, guid_bytes, 
                             hostID.length + time_bytes.length,
                             randID.length);
        } 
	catch(Exception e) 
	{
            System.err.println("Got exception creating GUID");
	    throw e;
        }
	
        guid = new Set_GUID(guid_bytes, len);
        hash_Table = new Hashtable();
	
        listeners = new Vector(1, 10);   
    }
    
    /**
    *  Return the Set_GUID that uniquely identifies this set from all
    *  other connected to the message space.  A Set_GUID provides
    *  a mechanism for identifing nodes.
    */
    public synchronized Set_GUID 
    get_GUID()
    {
        return guid;
    }
        

    /**
    *  Returns the url of the message space.
    */
    public synchronized String 
    get_Message_Space_URL() 
    {
        return message_Space_Url;
    }
    
    /**
    *  add a State to the list managed by this Set.
    */
    public void 
    add_Shared_State(
        State                                  shared_state)
    {
        synchronized(this)
        {
            hash_Table.put(shared_state.get_Name(), shared_state);
            shared_state.joined_Set(this);

            messenger.add_Shared_State(shared_state);
        }

        /*
         * lock the state that is waiting
         */
        synchronized(shared_state)
        {
            while(!shared_state.get_Initialized())
            {
                try
                {
                    shared_state.wait();
                }
                catch (Exception e) 
                { 
                   e.printStackTrace();
                }
            }
        }
    }
    
    /**
    *  Returns a reference to a State of the passed in name.  If the
    *  Set does not contain a State of that name null is returned.
    */
    public synchronized State 
    get_Shared_State(
        String                                 shared_state_name)
    {
        return (State) hash_Table.get(shared_state_name);
    }
    
    /**
    *  Instatiates and returns a Serial object that the given State
    *  can use for serializing its State.  This method gets a Serial
    *  object from its messenger.  It then packs in header information
    *  that the set needs for messaging. 
    *  @param shared_state a reference to the shared_state object that
    *  will use the new Serial object for serializing itself.
    */
    public Serial 
    new_Shared_State_Serial( 
        State                                 shared_state)
    {
	Serial serial;

	serial = messenger.new_Shared_State_Serial();
	pack_User_Set(serial, shared_state);

	return serial;
    }

    /**
    *  This method adds a header to the passed in serial object that 
    *  contains data for serializing the passed in shared_state.
    */
    public synchronized void 
    pack_User_Set(
        Serial                                  serial, 
	State                                   shared_state)
    {
        String state_name;
        int name_len[] = new int[1];
        char char_name[];
        byte guid_bytes[];
        int  len[] = new int[1];
	
        guid_bytes = guid.get_Bytes();
        len[0] = guid_bytes.length;
	
        state_name = shared_state.get_Name();
        char_name = state_name.toCharArray();
        
        name_len[0] = char_name.length;
		
        serial.pack_Int(len, 0, 1);
        serial.pack_Byte(guid_bytes, 0, len[0]);
        serial.pack_Int(name_len, 0, 1);
        serial.pack_Char(char_name, 0, name_len[0]);
    }

    /**
       This method is called when a State managed by this set is 
       called.  All Set_Event_Listeners registered with this
       Set are notified of the change.
     */
    public void 
    state_Change_Occurred(
        String                               state_name, 
        Set_GUID                             guid)
    {
        int ctr;
        Set_Event_Listener l;
        
        for(ctr = 0; ctr < listeners.size(); ctr++)
	{
	    l = (Set_Event_Listener) listeners.elementAt(ctr);
	    if(l != this) l.state_Change_Occurred(state_name, guid);
	}
    }
    
    /**
       Called by the internal messenger object when a new Set
       joins message space.
     */
    public synchronized void 
    set_Joined(
        Set_GUID                              guid)
    {
        int ctr;
        Set_Event_Listener l;
        
        for(ctr = 0; ctr < listeners.size(); ctr++)
	{
	    l = (Set_Event_Listener) listeners.elementAt(ctr);
	    if(l != this) l.set_Joined(guid);
	}
    }

    /**
       Called by the internal messenger object when a new
       Set leaves message space.
     */
    public synchronized void 
    set_Exited(
        Set_GUID                            guid)
    {
        int ctr;
        Set_Event_Listener l;
        
        for(ctr = 0; ctr < listeners.size(); ctr++)
	{
	    l = (Set_Event_Listener) listeners.elementAt(ctr);
	    if(l != this) l.set_Exited(guid);
	}
    }

    /**
       Called by the messaging object when any set in space adds
       a state.
     */
    public synchronized void 
    state_Added(
        String                               state_name, 
	Set_GUID                             source_set_guid)
    {
        int ctr;
        Set_Event_Listener l;
        
        for(ctr = 0; ctr < listeners.size(); ctr++)
	{
	    l = (Set_Event_Listener) listeners.elementAt(ctr);
	    if(l != this) l.state_Added(state_name, source_set_guid);
	}
    } 


    /**
       Called by the internal messaging object to notify the Set
       that new data has been received from message space.

       Called by the handler thread
     */
    public synchronized void 
    receive(
        Serial                              serial, 
        int                                 message_type, 
        long                                seq_num)
    {
        String                              state_name;
        int                                 state_name_len[] = new int[1];
        char                                char_name[];
        State                               state;
        int                                 guid_len[] = new int[1];
        byte                                guid_bytes[] = null;

        try {
	    serial.unpack_Int(guid_len, 0, 1);
	    
            guid_bytes = new byte[guid_len[0]];
            serial.unpack_Byte(guid_bytes, 0, guid_len[0]);
	    
            serial.unpack_Int(state_name_len, 0 , 1);
            char_name = new char[state_name_len[0]];
            
            serial.unpack_Char(char_name, 0, state_name_len[0]);
            state_name = new String(char_name);

            state = (State)hash_Table.get(state_name);

            if(state != null)
            {
                if(message_type == State.INITIALIZE ||
                   message_type == State.NAK_INITIALIZE)
                {
                    state.set_Initialized(true);
                }

                if(message_type != State.NAK_INITIALIZE)
                {
 	            state.state_Changed(
                        serial, 
                        State.FINAL,
                        new Set_GUID(guid_bytes, guid_len[0]));
                }
            }

            if(message_type == State.FINAL)
            {
	        state_Change_Occurred(
                    state_name, 
	            new Set_GUID(guid_bytes, guid_len[0]));
            }
        }
        catch(Exception lpe) 
	{
            lpe.printStackTrace();
	    System.err.println(
                "Lost precsion/buffer everrun in set header pack.");   
        }

    }

    /**
       Sends an event across message space notify all other Sets that
       a change request to a given State have been made.
       @param serial the serialized value of the changed State
       @param state_changed_type the message type.  Intermediate (Unrelaible)
       FINAL (reliable)
     */
    public void 
    set_State_Changed(
        Serial                               serial, 
        int                                  state_changed_type)
    {
        messenger.send(serial, state_changed_type);
    }
    
    /**
       Leave message space.
     */
    public void close()
    {
        messenger.close();
    }
}
