package cif.shared_state;

import java.lang.*;
import java.util.*;
import nexus.*;

/**
   An abstract class that provides the user with an interface to shared data. 
    
<blockquote>  

In the cif.shared_state package each process contains a local copy of 
all the abstract states in the message space.  States are added to each local
Set. Each State in a given set must have a unique String
associated with it that acts as its unique ID.  When a local copy of a state 
is changed the "pack()" method is called and the new value of the state
is serialized into a byte array.  This byte array is passed to the
Set associated with the State.  The Set propigates the new serialized value
of this state to all Sets in the message space.  When the Set receives
this message its list of States is checked for a unique id match.  
If the Set contains a state with the same unique id that state's "unpack()" 
method is called and passed the serialized byte array.  The unpack method 
sets the local State to the  serialized value thus syncronizing it with all
States of the same name in message space.
<BR>
All extended classe must implement:
public void unpack(Serial serial);
public void pack(Serial serial);
public int  get_State_Count();
<BR>
known implementations:
<BR>
State_Int, State_Float, State_Short, State_Long, State_Byte, State_Char,
State_Float
</blockquote> 


*@author John Bresnahan  
*/  
 
//***************************************************************************
//*   CIF_Shared_State
//*   ----------------
//***************************************************************************
public abstract class 
State
{
    protected String          shared_State_Name;
    protected Vector          callback_List;
    protected Set             shared_State_Set = null;

    protected boolean         initialized = false;
    
    /** FINAL: reliable message type */
    public static final int   FINAL = 0;
    /** INTERMEDIATE: unreliable message type */
    public static final int   INTERMEDIATE = 1;
    /** INITIALIZE: Message indicating that the current message is being
     used to initialize the state.  All messages that come in before this
    message are ignore.  Messages should never be sent with this type.*/
    public static final int   INITIALIZE = 2;
    /** INITIALIZE: Message indicating that the current message is being
     used to initialize the state.  All messages that come in before this
    message are ignore.  Messages should never be sent with this type.*/
    public static final int   NAK_INITIALIZE = 3;
   

    /**  
	 Create a shared state with a given name.  
	 @param name the name of the State.  The name must be
	 unique in regard to all other State names in the message space
    */ 
    public 
    State(
        String                             name)
    {
	shared_State_Name = new String(name);
	callback_List = new Vector(10, 10);
    }

    /**  
	 Create a shared state with a given name and register  
	 a callback interface.
	 @param name The name of the State.  The name must be
	 unique in regard to all other State names in the message space.
	 @param callback An object that implements the Callback_Interface 
	 interface.  This object will receive event notification when the 
	 value of the state changes.
    */ 
    public 
    State(
        String                              name,
	Callback_Interface                  callback)
    {
	shared_State_Name = new String(name);
	callback_List = new Vector(10, 10);

	add_Callback(callback);
    }


    /**
       Called by Set when the current state is added a set.
    */
    void  
    joined_Set(
        Set                                set)
    {
	shared_State_Set = set;
    }

    
    public synchronized void 
    set_Initialized(
        boolean                            b)
    {
	this.initialized = b;
        if(b)
        {
            this.notify();
        }
    }
  
    /*
     *
     */
    public synchronized boolean 
    get_Initialized()
    {
	return initialized;
    }

    /**
       Return the name of the State.  This name is unique to all States in
       a given set.
     */
    public String 
    get_Name()
    {
	return new String(shared_State_Name);
    }
       
    /**
       An abstract class that must be implemented by a child class.
       It returns the state count.  If zero is returned the system
       assumes the state has a null value.
     */
    public abstract int  get_State_Count();
    
    /**
       An abstract method that serializes the value of the shared
       state into the passed Serial object.  Unpack and pack must be 
       overridden by the child class.  They provide a mechanism for 
       syncronisation across the messaging wire.
     */
    public abstract void  pack_State(Serial   serial);
    
    /**
       An abstract class the unpacks the passed in serial object and sets 
       the current state to that value.  The serial format for each object
       is determined by the implementing class.
     */
    public abstract void  unpack_State(Serial serial) 
	throws LostPrecisionException,
	       BufferOverrunException;

    /**
       state_Changed is called by the Set object in which the associated state
       is a member.  It is in this method that the local state set receives
       notification that a State in the message space with the same 
       unique id has changed.  The method unpacks the new serialized value
       of the state and notifies all registered Callback_Interface objects
       of the change.
     */
    public synchronized void
    state_Changed(
        Serial                                serial, 
	int                                   state_changed_type,
        Set_GUID                              guid) 
	throws LostPrecisionException,
	       BufferOverrunException
    {
	int                                   ctr;
	Callback_Interface                    callback;

	unpack_State(serial);

	for(ctr = 0; ctr < callback_List.size(); ctr++)
	{
	    callback = (Callback_Interface)callback_List.elementAt(ctr);
	    callback.state_Changed(this, state_changed_type, guid);
	}
    }

    /**
       Add a Callback_Interface to the list that receive notification when 
       then value of the state changes.
     */
    public synchronized void 
    add_Callback(
        Callback_Interface                        callback)
    {
	callback_List.addElement(callback);
    }
    
    /**
       Remove a Callback_Interface from the list that receive notification
       of state changes.
     */
    public synchronized void 
    remove_Callback(
        Callback_Interface                      callback)
    {
	callback_List.removeElement(callback);
    }   
}
