package cif.shared_state.primative;

import java.lang.*;
import nexus.*;
import cif.shared_state.*;
/**
A shared double array implemention of the abstract State class.
*/
/*****************************************************************************
*  CIF.Shared_State.State_Double
*  ---------------------------
*
*   Wire format
*   0     = total length
*   1     = offset
*   2     = count
*   3-[2] = int array
*****************************************************************************/
public class State_Double extends State
{
    protected double     state_Value[] = null;
    protected int       max_Array_Size;
    
 
    /**
       Create a shared Double array.
       @param name Identifier of the State unique to the Set.
       @param size The length of the array.
     */
    public
        State_Double(String name, int size)
    {
	super(name);
        state_Value = new double[size];
	max_Array_Size = size;
    }

    /**
       Create a shared Double array.
       @param name Identifier of the State unique to the Set.
       @param size The length of the array.
       @param callback a callback interface that will be notified
       when the shared_state changes.
     */
    public 
        State_Double(String name,  
		    int size,
		  Callback_Interface callback)
    {
	super(name, callback);
        state_Value = new double[size];
	max_Array_Size = size;
    }

    /**
       Create a shared Double array.
       @param name Identifier of the State unique to the Set.
       @param init_val sets the initial value to this array.
       @param size The length of the array.
       @param callback a callback interface that will be notified
       when the shared_state changes.
     */
    public State_Double(String name,
                       double init_val[],
		       int size,
		       Callback_Interface callback)
    {
	super(name, callback);
        state_Value = new double[size];
	max_Array_Size = size;
	System.arraycopy(init_val, 0, state_Value, 0, size);
    }
    
    
    /*
     *
     */
    private boolean
	pack_User_State(Serial serial, 
			int dest_offset,
                        double src_state[], 
			int src_offset, 
			int count) 
			
    {
	int temp_int[] = new int[3];
	
	if(src_offset + count > max_Array_Size ||
	   dest_offset + count > max_Array_Size)  return false;
	
	temp_int[0] = max_Array_Size;
	temp_int[1] = dest_offset;
	temp_int[2] = count;
	
	serial.pack_Int(temp_int, 0, 3);

	// if there is no data to pack return.
	if(count == 0) return true;
        serial.pack_Double(src_state, src_offset, count);
	
	return true;
    }

    /**
       returns the length of the shared array.
    */
    public int 
	get_State_Count()
    { 
	return max_Array_Size;
    }
    
    /**
       Copies the current value of the shared_state into the passed in 
       array.
       
       @param dest_state array where the current value will be copied.
       @param dest_offset index into dest_state where coping short
       begin.
       @param src_offset index into the source array where coping should
       begin.
       @param count the number of items to copy.
    */
    public synchronized boolean
        get_State(double dest_state[], 
		  int dest_offset,
		  int src_offset, 
		  int count)
    {
	if(!initialized || src_offset+count > max_Array_Size) return false;
	
	System.arraycopy(state_Value, src_offset, 
			 dest_state, dest_offset, 
			 count);
	
	return true;
    }
    
    /**
       This method makes a request to set the value of the
       shared array to the passed in array.  If the request is 
       granted a callback event will come in from the Set
       to the State and set the the current value.  All Callback_Interfaces
       registered with this State will be notifiedof the change.
       
       @param dest_offset index into State array where the first value will
       be changed.
       @param src_state the array that the current value will be set to.
       @param src_offset the index into src_array where coping should begin.
       @param count the number of items in the array to change.
       @param state_changed_type the type of message request, FINAL=reliable, 
       INTERMEDIATE=unreliable.
    */
    public boolean          
	set_State_Changed(int dest_offset, 
                          double src_state[], 
			  int src_offset, 
			  int count,
			  int state_changed_type)
    {
	Serial serial;
	boolean b;
	
	serial = shared_State_Set.new_Shared_State_Serial(this);
	
	b = pack_User_State(serial, dest_offset, src_state, 
			    src_offset, count);
	
	if(!b) return false;

	shared_State_Set.set_State_Changed(serial, state_changed_type);
	
	return true;
    }
    
    
    /**
       Packs the value of the state into the passed in Serial object.
     */
    public synchronized void  
	pack_State(Serial serial)
    {
	pack_User_State(serial, 0, state_Value, 0, max_Array_Size);
    }
    
   
    /**
       Unpacks the passed in Serial objects into the current
       State value.
     */
    public synchronized void
	unpack_State(Serial serial) throws LostPrecisionException,
	                                   BufferOverrunException
    {
	int temp_int[] = new int[3];

	serial.unpack_Int(temp_int, 0, 3);

	// probable should throw an exception
	if(temp_int[0] > max_Array_Size)
	{
	    temp_int[0] = max_Array_Size;
	}

	if(max_Array_Size < temp_int[1] + temp_int[2])
	{
	    temp_int[2] = max_Array_Size - temp_int[1];
	}
        serial.unpack_Double(state_Value, temp_int[1], temp_int[2]);
    }
}

