
package nexus;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.io.InvalidClassException;


public class Startpoint
{
    private boolean				is_null;
    private Nexus				nexus;
    private OutgoingInterface			outgoing;
    byte[]					comm_descriptor_table;
    private boolean				copy_locally;
    private boolean				destroy_locally;
    private int					transform_id;
    private long				endpoint_id;
    private byte[]				liba;

    private static final int VERSION = 0;
    
    private static final int NEXUS_TRANSFORM_NONE = 0;
    private static final int LIBA_SIZE = 8;    
    private static final int FORMAT = DataConversion.NEXUS_DC_FORMAT_JAVA;
    
    /*
     * Startpoint()
     *
     * To instantiate a startpoint from an endpoint
     * (startpoint bind).
     */
    public
    Startpoint(
	Nexus					nexus,
	Endpoint				ep)
    {
	this.nexus = nexus;
	if (ep == null)
	{
	    this.is_null = true;
	}
	else
	{
	    ProtoInfo protoInfo;

	    this.is_null = false;

	    // get ep's protcol
	    protoInfo = ep.get_protocol();

	    // get ep's id
	    endpoint_id = ep.id();

	    this.comm_descriptor_table =
		ep.get_comm_descriptor_table();
	    this.outgoing =
		nexus.outgoing_instantiate(comm_descriptor_table);
	    this.copy_locally = false;
	    this.destroy_locally = false;
	    this.transform_id = NEXUS_TRANSFORM_NONE;
        
	    /* Initialize the liba byte array using the endpoint_id */
	    liba = new byte[LIBA_SIZE];
	    DataConversion.put_long(endpoint_id, liba, 0);
	}
    }
    /* Startpoint() */

    /*
     * Startpoint()
     *
     * Construct a startpoint from the passed buffer.
     */
    public
    Startpoint(
	Nexus					nexus,
	GetBuffer				buffer)
    throws BufferOverrunException,
	   LostPrecisionException,
	   InvalidClassException
    {
	this.nexus = nexus;

	/* get version and check it */
	byte					version;
	
	version = buffer.get_byte();
	if (version != (byte) Startpoint.VERSION)
	{
	    this.is_null = true;
		
	    throw new InvalidClassException(
		"Startpoint.Startpoint(buffer): " +
		"incompatible version in buffer, extraction aborted");
	}
	    
	/* get is_null */
	is_null = (buffer.get_byte() == 1) ? true : false;
    
	if (!is_null)
	{
	    /* get endpoint_id */
	    this.endpoint_id = buffer.get_long();
    
	    /* get copy_locally, destroy_locally, transform_id */
	    this.copy_locally = (buffer.get_byte() == 0) ? false : true;
	    this.destroy_locally = (buffer.get_byte() == 0) ? false : true;
	    this.transform_id = (int)buffer.get_byte();
    
	    /* get the liba */
	    this.liba = new byte[buffer.get_int()];
	    buffer.get_byte(this.liba, 0, this.liba.length);
        
	    /* get the comm_descriptor_table */
	    this.comm_descriptor_table = new byte[buffer.get_int()];

	    buffer.get_byte(this.comm_descriptor_table,
			    0,
			    this.comm_descriptor_table.length);

	    /* get transform_state */
	    if (this.transform_id != NEXUS_TRANSFORM_NONE)
	    {
		/* TODO: This needs to be filled in someday */
		System.out.println(
		    "\nNexusJava does not yet support transforms");
	    }
    
	    /* Instantiate the comm_descriptor_table to an OutgoingInterface */
	    this.outgoing = nexus.outgoing_instantiate(comm_descriptor_table);
	}
    }
    /* Startpoint() */

    /*
     * Startpoint()
     *
     * Construct a startpoint from the passed InputStream
     */
    public
    Startpoint(
	Nexus					nexus,
	InputStream				in,
	int					format)
    throws InvalidClassException
    {
	this.nexus = nexus;
	int int_size = DataConversion.sizeof_int(1, format);
	int long_size = DataConversion.sizeof_long(1, format);
	byte[] bytes = new byte[long_size];
    
	try
	{
	    /* get version and check it */
	    byte				version;
	    
	    in.read(bytes, 0, 1);
	    version = bytes[0];
	    if (version != (byte) Startpoint.VERSION)
	    {
		this.is_null = true;
		
		throw new InvalidClassException(
		    "Startpoint.Startpoint(buffer): " +
		    "incompatible version in buffer, extraction aborted");
	    }

	    /* read is_null */
	    in.read(bytes, 0, 1);
	    this.is_null =
		(bytes[0] == 1) ? true: false;
	
	    if (!is_null)
	    {
		/* read endpoint id */
		in.read(bytes);
		endpoint_id = DataConversion.get_long(bytes, 0, format);
        
		/* read copy locally, destroy locally, transform id */
		in.read(bytes, 0, 3);
		copy_locally = (bytes[0] == 1) ? true : false;
		destroy_locally = (bytes[1] == 1) ? true : false;
		transform_id = (int)bytes[2];
        
		/* read liba */
		in.read(bytes, 0, int_size);
		liba = new byte[DataConversion.get_int(bytes, 0, format)];
		in.read(liba);
        
		/* read comm descriptor table */
		in.read(bytes, 0, int_size);
		comm_descriptor_table =
		    new byte[DataConversion.get_int(bytes, 0, format)];
		in.read(comm_descriptor_table);
        
		/* read transform state */
		if (transform_id != NEXUS_TRANSFORM_NONE)
		{
		    /* TODO: this needs to be filled in eventually */
		    System.out.println(
			"\nNexusJava does not yet support transforms.");
		}
        
		/* Instantiate the comm_descriptor_table to an
                   OutgoingInterface */
		this.outgoing =
		    nexus.outgoing_instantiate(this.comm_descriptor_table);
	    }
	}
	catch (IOException x) {}
	catch (Exception x) { x.printStackTrace(); }
    }
    /* Startpoint() */   

    /*
     * put()
     *
     * Insert this startpoint into the passed buffer.
     * To get a startpoint from a buffer,
     * call the constructor that uses a buffer.
     */
    public
    void
    put(
	PutBuffer				buffer)
    throws BufferOverrunException
    {
	buffer.put_byte((byte) Startpoint.VERSION);
	
	if (is_null)
	{
	    buffer.put_byte((byte) 1);
	}
	else
	{
	    /* put is_null */
	    buffer.put_byte((byte) 0);

	    /* put endpoint id */
	    buffer.put_long(endpoint_id);
  
	    /* put copy_locally, destroy_locally, transform_id */
	    buffer.put_byte((byte)((copy_locally) ? 1 : 0));
	    buffer.put_byte((byte)((destroy_locally) ? 1 : 0));
	    buffer.put_byte((byte)transform_id);
  
	    /* put the liba */
	    buffer.put_int(liba.length);
	    buffer.put_byte(liba, 0, liba.length);
  
	    /* put the comm_descriptor_table */
	    buffer.put_int(comm_descriptor_table.length);
	    buffer.put_byte(comm_descriptor_table,
			    0, comm_descriptor_table.length);
  	    
	    /* put transform_state */
	    if (transform_id != NEXUS_TRANSFORM_NONE)
	    {
		/* TODO: This needs to be filled in someday */
		System.out.println(
		    "\nNexusJava does not yet support transforms");
	    }
	}
    }
    /* put() */

    /*
     * write()
     *
     * Write this sp into the passed OutputStream.
     * To get a sp from an InputStream, call 
     * the constructor that uses an InputStream.
     */
    public
    void
    write(
	OutputStream				out)
    {
	byte[] bytes = new byte[8];// TODO: should be LARGEST ULONG 
    
	try
	{
	    /* write version number */
	    bytes[0] = (byte) Startpoint.VERSION;
	    out.write(bytes, 0, 1);
	    
	    if (is_null)
	    {
		/* is_null==1 means it's a null startpoint */
		bytes[0] = 1;
		out.write(bytes, 0, 1);
	    }
	    else
	    {
		bytes[0] = 0;
		out.write(bytes, 0, 1);
        
		/* write endpoint_id */
		DataConversion.put_long(endpoint_id, bytes, 0);
		out.write(bytes);		
        
		/* write copy locally, destroy locally, transform id */
		bytes[0] = (byte)((copy_locally) ? 1: 0);
		bytes[1] = (byte)((destroy_locally) ? 1: 0);
		bytes[2] = (byte)transform_id;
		out.write(bytes, 0, 3);
        
		/* write liba */
		DataConversion.put_int(liba.length, bytes, 0);
		out.write(bytes, 0, 4);
		out.write(liba);
        
		/* write comm descriptor table */
		DataConversion.put_int(comm_descriptor_table.length, bytes, 0);
		out.write(bytes, 0, 4);
		out.write(comm_descriptor_table);
        
		/* put transform state */
		if (transform_id != NEXUS_TRANSFORM_NONE)
		{
		    /* TODO: this needs to be filled in eventually */
		    System.out.println(
			"\nNexusJava does not yet support transforms.");
		}
	    }
	}
	catch (IOException x){}
	catch (Exception x) { x.printStackTrace(); }
    }
    /* write */		

    /*
     * sizeof()
     *
     * Compute the size of this startpoint's
     * byte array representation.
     */
    public
    int
    sizeof()
    {
	int size = 0;
    
	/* sizeof version */
	size += 1;
    
	/* sizeof is_null */
	size += 1;
    
	if (!is_null)
	{
	    /* size of endpoint id */
	    size += DataConversion.sizeof_long(1);
    
	    /* sizeof copy_locally, destroy_locally, transform_id */
	    size += 3;
    
	    /* sizeof liba */
	    size += DataConversion.sizeof_int(1);
	    size += liba.length;
    
	    /* sizeof comm_descriptor_table */
	    size += DataConversion.sizeof_int(1);
	    size += comm_descriptor_table.length;
    
	    /* sizeof transform_state */
	    if (transform_id != NEXUS_TRANSFORM_NONE)
	    {
		/* TODO: This needs to be filled in someday */
		System.out.println(
		    "\nNexusJava does not yet support transforms");
	    }
	}

	return size;
    }
    /* sizeof() */

    /*
     * set_null()
     */
    public
    void
    set_null()
    {
	is_null = true;
    }
    /* set_null */
    
    /*
     * is_null()
     */
    public
    boolean
    is_null()
    {
	return is_null;
    }
    /* is_null */
    
    /*
     * liba()
     */
    public
    byte[]
    liba()
    {
	return liba;
    }
    /* liba() */
    
    /*
     * destroy()
     */
    public
    void
    destroy()
    {
	outgoing.destroy();
	outgoing = null;
    }
    /* destroy() */

    /*
     * outgoing()
     */
    public
    OutgoingInterface
    outgoing()
    {
	return outgoing;
    }
    /* outgoing() */
    
} /* class Startpoint */

