package nexus;

import nexus.totem.*;
import java.util.*;
import java.net.*;
 
public class TotemProto
  implements ProtoInterface
{
  public static final int PROTOCOL_VERSION = Nexus.BUFFER_VERSION + 2;
  public static final int MI_PROTO_VERSION = 2;

  private Nexus        nexus;
  private Hashtable    readerTable;
  private PG_Inf       sendPG;
  private boolean      sendConnected = false;

  private String       daemonHostname;
  private int          daemonPort;

  /*
   * TotemProto()
   */
  public TotemProto(Nexus nexus) 
  {
    this.nexus = nexus;

    // as messages are received by the reader
    // they are placed on the queue to be handled
    // by the handler
    readerTable = new Hashtable();

    daemonPort = 4589;
    daemonHostname = System.getProperty("GLOBUS_TOTEM_DAEMON");
    try
    {
       if(daemonHostname == null) 
       {
	  daemonHostname = InetAddress.getLocalHost().getHostName();
       }
    }
    catch (Exception e)
    {
       daemonHostname = "localhost";
    }

    System.out.println("GLOBUS_TOTEM_DAEMON = " + daemonHostname);
    sendPG = new PG_Inf();
  } /* TotemProto() */

  /*
   * get_my_mi_proto()
   * -----------------
   * called by endpoint
   */
  public byte[] get_my_mi_proto(ProtoInfo protoInfo,
                                long endpointID) 
  {
    TotemInfo info; 
    byte[] miProto;
    Queue        msgQueue = new Queue();

    // get user-specified connection information
    info = (TotemInfo)protoInfo;

    // build mi_proto array
    miProto = new byte[DataConversion.sizeof_long(1) + 2];
    miProto[0] = (byte)(DataConversion.NEXUS_DC_FORMAT_JAVA);
    miProto[1] = (byte)(TotemProto.PROTOCOL_VERSION);
    DataConversion.put_long((long)(info.pgid), miProto, 2);


    info.daemonHost = daemonHostname;
    info.daemonPort = daemonPort;
    TotemReader reader = new TotemReader(info, endpointID, this, msgQueue, nexus);
    readerTable.put(new Long(endpointID), reader);
    NexusThread rThread = new NexusThread(nexus, reader, "totem_reader");
    rThread.start();

    return miProto;
  } /* get_my_mi_proto() */

  /*
   * construct_from_mi_proto()
   * called from startpoint bind, or copy, or get
   */
  public OutgoingInterface construct_from_mi_proto(byte[] mi_proto,
                                                   int start,
                                                   int size,
                                                   String node_name,
                                                   boolean[] is_valid) 
  {
    TotemOutgoing   outgoing;
    int             pgid;
    String          hostname;

    if(!sendConnected)
    {
       try
       {
          sendPG.ProcessConnect(daemonHostname, daemonPort);
          sendConnected = true;
       }
       catch (Exception e)
       {
           System.out.println("unable to get totem daemon hostname");
       }
    }
    // get pgid - process group to send outgoing message to
    try 
    {
      pgid = (int)(DataConversion.get_long(mi_proto, start + 2,
                                           (int)(mi_proto[start])));
    }
    catch (LostPrecisionException lpX) 
    {
      pgid = 0;
      System.err.println("TotemProto.construct_from_mi_proto(): "
                         + "Lost precision on pgid conversion.  "
                         + "Using pgid = " + pgid);
    }
   
    return new TotemOutgoing(this, pgid);
  } /* construct_from_mi_proto() */ 

  /*
   * shutdown()
   */ 
  public void shutdown()
  {
    Enumeration rEnum = readerTable.elements();

    while(rEnum.hasMoreElements())
    {
       TotemReader r = (TotemReader)rEnum.nextElement();
       r.shutdown();
    }
    if(sendConnected)
    {
        sendPG.close();
    }
  } /* shutdown() */

  /*
   * send_rsr()
   *
   * sends a message to a Process Group
   *  int     length of complete message
   *  int     message type can be AGREED, SAFE
   *  InfId[] an array defining process groups to send to
   *  byte[]  data to be sent
   *  returns an integer that is used to flag any error
   */
  public void send_rsr(byte[] buffer,
                       int pgid) 
  {

    InfId[] infId = new InfId[1];
    infId[0] = new InfId(pgid);

    if (sendPG.ProcessSend(buffer.length, PG_Inf.AGREED, infId, buffer) == -1) 
    {
      System.err.println("TotemProto.send_rsr(): "
                         + "Error occurred while sending message");
    }

  } /* send_rsr() */

    public void destroy_my_mi_proto(Endpoint ep, byte mi_proto[])
    {

       long epID = ep.id();
       TotemReader reader = (TotemReader)readerTable.remove(new Long(epID));
       if(reader != null)
       {
          reader.shutdown();
       }
       else
       {
           System.err.println("Error in totem proto");
       }
       
      // called from endpoint destroy 

    }

} /* TotemProto */


/*
 * TotemReader
 *   receives messages from process group layer
 */
class TotemReader
  implements Runnable 
{

  private PG_Inf     pg;
  private TotemProto totemProto;
  private Queue      msgQueue;
  private Nexus      nexus;
  private volatile boolean    done = false;
  private TotemInfo  totemInfo;
  private boolean    connected=false;
  private long       epID;
  private boolean    dieing = false;
  /*
   * TotemReader()
   */
  public TotemReader(TotemInfo  info,
		     long endpointID,
                     TotemProto totem,
                     Queue msgQueue,
                     Nexus nexus) 
  {
    this.totemInfo = info;
    this.totemProto = totem;
    this.msgQueue = msgQueue;
    this.nexus = nexus;
    this.epID = endpointID;

    pg = new PG_Inf();


   if(pg.ProcessConnect(info.daemonHost, info.daemonPort) != -1 &&
       pg.ProcessJoin(info.pgid) != -1)
    {
        connected = true;
    }
    else
    {
       System.err.println("failed to connect to totem server");
    }
  } /* TotemReader() */

  /*
   * run() 
   */
  public void run() 
  {
      nexus.add_thread(Thread.currentThread()); 

      while (!nexus.is_destroying() && !done)  
      {
          // don't try to read if connection failed
          if (connected) 
          {
	      int __x = pg.ProcessReceive();
              if (__x != -1) 
	      {
		  debug_print("about to call receive");
                  this.receive();
		  debug_print("returned from receive");
              }
             else 
	     {
                  System.err.println("TotemReader.run(): Error receiving message.");
             }
          }
       }

    nexus.remove_thread(Thread.currentThread());
 
  } /* run() */

  /*
   * receive()
   */
  private synchronized void receive() 
  {
    debug_print("start receive()");
    if ((pg.infh.type == PG_Inf.AGREED || pg.infh.type == PG_Inf.SAFE) 
	&& !dieing) 
    {    
        debug_print("Regular type message");
        byte[] b = new byte[pg.infh.msgLength()];
        System.arraycopy(pg.data, 0, b, 0, pg.data.length);

	debug_print("pre buffer dispatch:"+epID);
        nexus.buffer_dispatch(b, epID);
	debug_print("post buffer dispatch");
    }
    else if(pg.infh.type == PG_Inf.LEAVE_ANS
	    && pg.sender.pid == pg.my_id.pid)
    {
        debug_print("Got leave confirmation");
        done = true;
	notify();
    }

    debug_print("end receive()");
  } /* receive() */

  /*
   * shutdown()
   */
  public synchronized void shutdown() 
  {
    debug_print("Leaving group:"+totemInfo.pgid+":"+Thread.currentThread());
    dieing = true;
    pg.ProcessLeave(totemInfo.pgid);
    while(!done) 
    {
    debug_print("before wait");
       try{wait();}catch (Exception e) {}
    debug_print("after wait");
    }
    pg.close();
    debug_print("%%%%%end shutdown");
  } /* shutdown() */

  private static void debug_print(String str)
  {
     //System.out.println("t:"+Thread.currentThread()+":"+str);
  }
} /* TotemReader */
     
