
package nexus;

import java.net.*;
import java.io.*;

/*
 * UdpOutgoing
 *
 * Don't need to check IP address for multicast range.  Any socket
 * may send messages to a multicast group.  There is no need for the 
 * socket to be part of the group.
 */
public class UdpOutgoing
    implements OutgoingInterface
{
    private Nexus				nexus;
    private UdpProto				udp_proto;
    private String                              hostname;
    private int                                 port;

    private boolean				is_open;
    private boolean				shuttingDown;
    private DatagramSocket			socket;

    private UdpInfo				info;
    private InetAddress				address;
    private static int				seqNum = 0;

    private byte                                ttl;
    private boolean                             multiSender = true;   

    /*
     * UdpOutgoing()
     */
    public
    UdpOutgoing(
	Nexus					nexus,
	UdpProto				udp_proto,
	String					hostname,
	int					port,
	byte					ttl,
	boolean                                 multi_sender)
    {
        this.nexus = nexus;
        this.udp_proto = udp_proto;
        this.hostname = hostname;
        this.port = port;
        this.ttl = ttl;
	this.multiSender = multi_sender;

	try
	{
	    this.address = InetAddress.getByName(hostname);
	}
	catch (UnknownHostException uhe)
	{
	    this.address = null;
	}

	this.is_open = false;
	this.shuttingDown = false;

        udp_proto.add_outgoing(this);
    }
    /* UdpOutgoing() */

    /*
     * UdpOutgoing.get_Protocol_Version()
     */
    public
    int
    get_Protocol_Version()
    {
	return UdpProto.PROTOCOL_VERSION;
    }
    /* UdpOutgoing.get_Protocol_Version() */
    

    /*
     * send_rsr()
     */
    public synchronized
    void
    send_rsr(
	byte[]					buffer)
    {
        byte					pack[];
        int					size;
        int					totalLength;
        int					bytesSent;
	boolean                                 frag0;

	Nexus.debug_println(2, "UdpOutgoing.send_rsr(): entering");

	/*
	 *  If outgoing was given an invalid hostname
	 */
	if(address == null) 
	{
	    System.err.println("UdpOutgoing.send_rsr() : null address.");

	    return;
	}
        // open the socket if necessary
        if (!is_open)
        {
	    Nexus.debug_println(2, "UdpOutgoing.send_rsr(): " +
				"opening socket to "+hostname+":"+port);

	    // If multicast request, create a MulticastSocket
	    if (address.isMulticastAddress())
	    {
		try
		{
		    socket = new MulticastSocket();
		    ((MulticastSocket)socket).setTTL(ttl);
		    //TODO: test this - joinGroup() necessary?
		    //      ((MulticastSocket)socket).joinGroup(address);
		    is_open = true;
		}
		catch (IOException ioX)
		{
		    System.err.println("UdpOutgoing.send_rsr(): unable to "
				       + "create multicast socket for "
				       + address + ":" + port);
		}
	    }
	    // if not multicast address create a datagram socket
	    else
	    { 
		try
		{
		    socket = new DatagramSocket();
		    is_open = true;
		}
		catch (SocketException hostX)
		{
		    System.err.println("UdpOutgoing.send_rsr(): unable to "
				       + "create datagram socket for "
				       + hostname + ":" + port);
		}
	    }
	}

	/*  if single sender situation it is ok for the message to be
	 *  fragmented, however if we are dealing with multiple senders
	 *  a error message is printed and the methodreturns.
	 */
	if (multiSender &&
	    buffer.length > UdpProto.UDP_MAX_BODY_SIZE)
	{
	    System.err.println(
		"UdpOutgoing.send_rsr(): " +
		"messsage to large for multiple sender protocol");

	    return;	    
	}
	
        totalLength = buffer.length;
	bytesSent = 0;
	frag0 = true;

	while(bytesSent < totalLength)
	{
	    if (totalLength - bytesSent < UdpProto.UDP_MAX_BODY_SIZE)
	    {
		size = totalLength - bytesSent; 
	    }
	    else
	    {
		size = UdpProto.UDP_MAX_BODY_SIZE;
	    }

	    pack = UdpMessage.packHeader(frag0, seqNum, size);

	    frag0 = false;
	    System.arraycopy(buffer, bytesSent,
			     pack, UdpProto.UDP_HDR_SIZE,
			     size);

	    this.send(pack, pack.length);

	    bytesSent += size;
	    seqNum++; 
	}

    }
    /* send_rsr() */
    
    /*
     * send()
     */
    private synchronized
    void
    send(
	byte[]					buffer,
	int					size)
    {
        DatagramPacket p;

        try
        {
            p = new DatagramPacket(buffer, size, address, port);
            socket.send(p);
        }
        catch (IOException ioX) 
        {
	    try
	    {
		if (socket instanceof MulticastSocket)
		{
		    ((MulticastSocket)socket).leaveGroup(address);
		}
	    }
	    catch (IOException _ioX)
	    {
		System.err.println("UdpOutgoing.send()  ioexecption");
	    }
	  
	    System.err.println("UdpOutgoing.send(): send failed");
	    socket.close();
	    is_open = false;
        }
    }
    /* send() */
    
    /*
     * UdpOutgoing.destroy()
     */
    public
    void
    destroy()
    {
	shutdown();
    }
    /* UdpOutgoing.destroy() */

    
    /*
     * shutdown()
     */
    public synchronized
    void
    shutdown()
    {
	Nexus.debug_println(2, "UdpOutgoing.shutdown()");

	if (!shuttingDown)
	{
            shuttingDown = true;
            if (is_open)
            {
                socket.close();
                is_open = false;
            }
	    
	    udp_proto.remove_outgoing(this);
	}
	
	Nexus.debug_println(2, "UdpOutgoing.shutdown(): " +
			    "Removed UdpOutgoing - shutdown");
    }
    /* shutdown() */
}
/* UdpOutgoing */
