package nexus;

import java.util.*;
import java.net.*;
import java.io.*;

/*
 * UdpProto
 */
public class UdpProto
    implements ProtoInterface
{
  // class statics
    public static final int PROTOCOL_VERSION = Nexus.BUFFER_VERSION + 1;
    public static final int MI_PROTO_VERSION = 1;
    
    public static final int UDP_HDR_SIZE = 4;
    public static final int UDP_MAX_TRANSMISSION_UNIT = 8192;
    public static final int UDP_MAX_BODY_SIZE =
    UDP_MAX_TRANSMISSION_UNIT - UDP_HDR_SIZE;

    // object vars
    private Nexus         nexus;
    private Queue         packet_queue;
    private UdpHandler    handler;
    private Vector        reader_vector;
    private boolean       reader_vector_done = false;
    private Vector        outgoing_vector;
    private boolean       outgoing_vector_done = false;
    private byte[]        default_mi_proto;

    /*
     * UdpProto()
     */
    public
    UdpProto(
	Nexus					nexus)
    {
	NexusThread				thread;
	Properties				properties;
	String					property;

	this.nexus = nexus;

	/*
	 * Create a message queue.
	 * As messages are received by the various threads watching
	 * the various InputStreams, those threads will put the messages
	 * onto the message queue.  A separate handler thread will deal
	 * with them.
	 */
	packet_queue = new Queue();

	outgoing_vector = new Vector(5, 5);
	reader_vector = new Vector(5, 5);

 
	/*
	 * Start a handler thread.  This thread will watch the
	 * packet_queue for incoming packets, and invoke the
	 * handlers for them.
	 */
	handler = new UdpHandler(nexus, this, packet_queue);
	thread = new NexusThread(nexus, handler, "udp_handler");
	thread.start();
    }
    /* UdpProto() */

    /*
     * shutdown()
     */
    public
    void
    shutdown()
    {
	/* Wake up the handler thread */
	packet_queue.awaken_all_waiting_threads();

	/* Shut down outgoing connections */
	shutdown_outgoings();

	/* Shut down incoming connections */
	shutdown_readers();
    }
    /* shutdown() */

    /*
     * add_reader()
     */
    public synchronized
    void
    add_reader(
	UdpReader				reader)
    {
	reader_vector.addElement(reader);
    }
    /* add_reader() */

    /*
     * remove_reader()
     */
    public synchronized
    void
    remove_reader(
	UdpReader				reader)
    {
	/* removeElement() will change the Enumeration of that Vector */
	if (!reader_vector_done)
	{
	    reader_vector.removeElement(reader);
	}
    }
    /* remove_reader() */
				    
    /*
     * get_final_readers()
     */
    private synchronized
    Enumeration
    get_final_readers()
    {
	reader_vector_done = true;
	return (reader_vector.elements());
    }
    /* get_final_readers() */

    /*
     * shutdown_readers()
     */
    private
    void
    shutdown_readers()
    {
	UdpReader reader;

	Nexus.debug_println(2, "UdpProto.shutdown_readers(): entering");
	for (Enumeration e = get_final_readers(); e.hasMoreElements(); )
	{
	    reader = (UdpReader) e.nextElement();
	    reader.shutdown();
	}
	Nexus.debug_println(2, "UdpProto.shutdown_readers(): exiting");
    }
    /* shutdown_readers() */

    /*
     * add_outgoing()
     */
    public synchronized
    void
    add_outgoing(
	UdpOutgoing				o)
    {
	outgoing_vector.addElement(o);
    }
    /* add_outgoing() */
   
    /*
     * remove_outgoing()
     */
    public synchronized
    void
    remove_outgoing(
	UdpOutgoing				o)
    {
	/* removeElement() will change the Enumeration of that Vector */
	if (!outgoing_vector_done) outgoing_vector.removeElement(o);
    }
    /* remove_outgoing() */

    /*
     * get_final_outgoings()
     */
    public synchronized
    Enumeration
    get_final_outgoings()
    {
	outgoing_vector_done = true;
	return (outgoing_vector.elements());
    }
    /* get_final_outgoings() */

    /*
     * shutdown_outgoings()
     */
    private
    void
    shutdown_outgoings()
    {
	UdpOutgoing				o;
	Nexus.debug_println(2, "UdpProto.shutdown_outgoings(): entering");
	for (Enumeration e = get_final_outgoings(); e.hasMoreElements(); )
	{
	    o = (UdpOutgoing) e.nextElement();
	    o.shutdown();
	}
	Nexus.debug_println(2, "UdpProto.shutdown_outgoings(): exiting");
    }
    /* shutdown_outgoings() */
	 
    /*
     * get_my_mi_proto()
     *
     * Construct a byte array communications descriptor (mi_proto).
     */
    public
    byte[]
    get_my_mi_proto(
	ProtoInfo				proto_info,
	long					endpointID)
    {
	InetAddress				inet_addr;
	DatagramSocket				socket;
	UdpReader				reader;
	UdpInfo					udp_info;
	byte[]					mi_proto;
	String					str_addr;
	boolean					unicast;
	String                                  address;
	boolean                                 multi_sender;
	boolean                                 loopback = true;
	int                                     fd = 0;
	
		
	// get user-specified connection information
	if (proto_info != null)
	{
	    udp_info = (UdpInfo) proto_info;
	}
	else
	{
	    udp_info = new UdpInfo();
	}

        /*
	 * Initialize my hostname
	 */
	if(udp_info.hostname == null)
	{
	  try
          {
	     udp_info.hostname = InetAddress.getLocalHost().getHostName();
	  }
	  catch (UnknownHostException e)
	  {
	     udp_info.hostname = new String("localhost");
	  }
        }

	multi_sender = udp_info.multiSender;
	loopback = udp_info.loopback;
	
	 try
	 {
	     inet_addr = InetAddress.getByName(udp_info.hostname);
	     address = inet_addr.getHostAddress();
	 }
	 catch (UnknownHostException e)
	 {
	    System.err.println(
	    "UdpInfo.UdpInfo(): Could not resolve address for" +
	    udp_info.hostname);
		
	    return mi_proto = new byte[1];
	 }
	    
	 // if multicast address
	 if (inet_addr.isMulticastAddress())
	 {
	    unicast = false;
 
	    // Multiple multicast readers with the same address 
	    // and port are acceptable - they will have unique
	    // endpoint id's (allows multiple multicast
	    // endpoints to exist in the same context with the
	    // same communication descriptors).

            // open socket
	    socket = null;
	    try
	    {
		    
	       // create MulticastSocket according to UdpInfo
	       socket = new MulticastSocket(udp_info.port);

	       // join multicast group
	       ((MulticastSocket)socket).setTTL(udp_info.ttl);
	       ((MulticastSocket)socket).joinGroup(inet_addr);
	    }
	    catch (IOException ioX)
	    {
	       if (socket != null)
	       {
	   	  try
		  {
		     socket.close();
		  } 
		  catch(Exception _e)
		  {
		     System.err.println(
			    "UdpProto: error closing socket");
		  }
	       }
		    
	       System.err.println(
		"UdpProto: error creating incoming multicast socket " +
		"for " + udp_info.hostname + ":" + udp_info.port);
		    
	         return mi_proto = new byte[1];
	    }

            // create reader to manage this incoming
	    reader = new UdpReader(nexus, this, socket, endpointID);
	    this.add_reader(reader);
	    Thread thread = new NexusThread(nexus, reader, "");
	    thread.start();
        }
        // user-specified unicast address
        else
	{
	   unicast = true;
	       

	   // TODO 
	   // Need to check UdpInfo against existing UdpReaders.
	   // If a reader with the same unicast address and port
	   // already exists, it should be returned (it would be
	   // impossible to open another on the same port).
	   try
	   {
	       InetAddress inet_temp = 
	        	InetAddress.getByName(udp_info.hostname);
		    
	       address = inet_temp.getHostAddress();

	       // create DatagramSocket according to UdpInfo
	       socket = new DatagramSocket(udp_info.port,
	         				inet_temp);

	       // create reader to manage this incoming
	       reader = new UdpReader(nexus, this, socket,
				   ProtoInterface.USE_BUFFER_EP_ID);
	       this.add_reader(reader);
	       Thread thread = new NexusThread(nexus, reader, "");
	       thread.start();
           }
	   catch (Exception sockX)
	   {
	       System.err.println("UdpProto: Couldn't create incoming "
				       + "connection for " + udp_info.hostname
				       + ":" + udp_info.port);
	       return mi_proto = new byte[1];
	   }
	}

	/* Java put; 32bit BE and Java have same int size and format */

	// place file descriptor number, port, and TTL in proto

	int port = socket.getLocalPort();
	int ttl;
	if (socket instanceof MulticastSocket)
	{
	    try
	    {
		ttl = ((MulticastSocket)socket).getTTL();
	    }
	    catch (IOException ioX)
	    {
		ttl = UdpInfo.RESTRICT_TO_SUBNET;
	    }
	}
	else
	{
	    ttl = UdpInfo.RESTRICT_TO_SUBNET;
	}

	// IP adress as String to bytes
	byte[] ip_addr = address.getBytes();
	mi_proto = new byte[ip_addr.length + 11];
	// ********************************************
	// what is fd:  
	// -----------
	// ********************************************
	// *  Java and 32bit BE havae same format
	// *  1) set byte 0     = VERSION
	// *  2) set bytes 3-6  = port
	// *  3) set bytes 1-4  = fd (overwrite first 2 of port)
	// ********************************************
	mi_proto[0] = UdpProto.MI_PROTO_VERSION;
	DataConversion.put_int(port, mi_proto, 3);
	DataConversion.put_int(fd, mi_proto, 1);
	mi_proto[7] = (byte)ttl;
	if(loopback)       mi_proto[8] = (byte)1;
	else               mi_proto[8] = (byte)0;
	if(multi_sender)   mi_proto[9] = (byte)1;
	else               mi_proto[9] = (byte)0;
	System.arraycopy(ip_addr, 0, mi_proto, 10, ip_addr.length);
	mi_proto[ip_addr.length + 10] = 0;

	// save default unicast mi_proto
	if (unicast && (default_mi_proto == null)) 
	{
	    default_mi_proto = mi_proto;
	}

	return (mi_proto);

    }
    /* get_my_mi_proto() */


    /*
     * construct_from_mi_proto()
     */
    public
    OutgoingInterface
    construct_from_mi_proto(
	byte[]					mi_proto,
	int					start,
	int					size,
	String					node_name,
	boolean[]				is_valid)
    {
	OutgoingInterface			outgoing;
	int					port;
	byte					ttl;
	String					hostname;
	int					length;
	boolean                                 multi_sender = true;
	boolean                                 loopback = true;

	if(mi_proto[start] != UdpProto.MI_PROTO_VERSION)
	{
            throw new Error("UdpProto.construct_from_mi_proto():" +
            	             "incompatible protocol version");
	}

	try
	{
	    byte pb[] = new byte[4];
	    System.arraycopy(mi_proto, start+3, pb, 0, 4);
	    pb[0] = (byte) 0;
	    pb[1] = (byte) 0;
	    port = DataConversion.get_int(
				  pb,
			          0,
				  DataConversion.NEXUS_DC_FORMAT_32BIT_BE);
	    ttl = (byte) mi_proto[start+8];
	    if(mi_proto[start+9] != 0) loopback = true;
	    else                       loopback = false;
	    if(mi_proto[start+10] !=0) multi_sender = true;
	    else                       multi_sender = false;
	} 
	catch (Exception x)
	{
	    System.err.println("LostPrecisionException occured " +
			       "when constructing OutgoingInterface");
	    port = -1;
	    ttl = -1;
	}

	// get hostname/ip address
	start += 10;  // skip fd, port, ttl
	for (length = 0; mi_proto[start+length] != 0; length++);
	if (length == 0)
	{
	    hostname = node_name;
	}
	else
	{
	    hostname = new String(mi_proto, start, length);
	}

	// do something with multiSender and loopback
	outgoing = new UdpOutgoing(nexus, this, hostname, 
				   port, ttl, multi_sender);
	is_valid[0] = true;

	return (outgoing);
    }
    /* construct_from_mi_proto() */

    /*
     * get_hostname()
     */
    protected String get_hostname()
    {
	return (new String(""));
    } /* get_hostname() */
    
    /*
     * enqueue_packet()
     */
    public void enqueue_packet(UdpPacket packet)
    {
	packet_queue.add(packet);
    }

    public void destroy_my_mi_proto(Endpoint ep, byte mi_proto[])
    {
    }

    /* enqueue_packet() */
}
/* UdpProto() */

/*
 * UdpReader
 */
class UdpReader implements Runnable
{
    private Nexus				nexus;
    private UdpProto				udp_proto;
    private DatagramSocket			socket;
    private boolean				closing_reader = false;
    private boolean				done = false;
    private long				endpointID;
    
    /*
     * UdpReader()
     */
    public
    UdpReader(
	Nexus					nexus,
	UdpProto				udp_proto,
	DatagramSocket				socket,
	long					endpointID)
    {
	this.nexus = nexus;
	this.udp_proto = udp_proto;
	this.socket = socket;
	this.endpointID = endpointID;

    }
    /* UdpReader() */

    /*
     * run()
     */
    public
    void
    run()
    {

	nexus.add_thread(Thread.currentThread());

	while (!nexus.is_destroying() && !done)
	{
	    try
	    {
		DatagramPacket packet = new DatagramPacket(
		    new byte[UdpProto.UDP_MAX_TRANSMISSION_UNIT],
		    UdpProto.UDP_MAX_TRANSMISSION_UNIT);
       
		socket.receive(packet);  
		udp_proto.enqueue_packet(new UdpPacket(packet, endpointID));
	    }
	    catch (Exception e)
	    {
		if (!closing_reader)
		{
		    System.err.println("ERROR: Reader.run(): " +
				       "Exception while reading: " + e);
		    e.printStackTrace();
		}
		done = true;
		continue;
	    }
	}
	this.shutdown();
	nexus.remove_thread(Thread.currentThread());
    }
    /* run() */

    /*
     * shutdown()
     */
    public synchronized
    void
    shutdown()
    {
	if (!closing_reader)
	{
	    closing_reader = true;
	    socket.close();
	}
    }
    /* shutdown() */
    
}
/* UdpReader */


/*
 * class UdpHandler
 */
class UdpHandler implements Runnable
{
    private Nexus				nexus;
    private UdpProto				udp_proto;
    private Queue				packet_queue;
    private Vector				msgTable;

    /*
     * UdpHandler()
     */
    public
    UdpHandler(
	Nexus					nexus,
	UdpProto				udp_proto,
	Queue					packet_queue)
    {
	this.nexus = nexus;
	this.udp_proto = udp_proto;
	this.packet_queue = packet_queue;

	msgTable = new Vector(10, 10);
    }
    /* UdpHandler() */

    /*
     * run()
     */
    public
    void
    run()
    {
	handle_messages();
	Nexus.debug_println(1, "UdpHandler.run(): " +
			    "handler thread terminating");
    }
    /* run() */
    
    /*
     * handle_messages()
     */
    private
    void
    handle_messages()
    {
	UdpPacket				inPack;
	byte[]					msg;
	int					version;
	boolean					frag0;
	int					seq_num;
	long					msg_total_length;
	UdpMessage				udpMessage;
	boolean					killFlag;

	while ((inPack = (UdpPacket) packet_queue.remove_with_wait()) != null)
	{
	    msg = inPack.getData();
	    version = (int) msg[0];
	    if (version != UdpProto.PROTOCOL_VERSION)
	    {
		throw new Error("UdpProto.Reader.run(): " +
				"incompatible protocol version");
	    }

	    frag0 = (((int) msg[1] & 0x80) != 0) ? true : false;
	    seq_num = ((int) msg[1] & 0x7f) << 16;
	    seq_num |= ((int) msg[2] << 8);
	    seq_num |= ((int) msg[3]);

	    /* if new packet */
	    if(frag0)
	    {
		version = (int)(msg[4]);
		int format = (int)(msg[5]);

		try
		{
		    msg_total_length = DataConversion.get_long(msg, 6, format);
		}
		catch (LostPrecisionException lpX)
		{
		    msg_total_length = -1;
		}

		udpMessage = new UdpMessage((int)(msg_total_length),
					    seq_num, 
					    inPack,
					    nexus);

		msgTable.addElement(udpMessage);
	    }

	    udpMessage = findNode(inPack, seq_num);
	    if(udpMessage != null)
	    {
		killFlag = udpMessage.addBytes(msg,
					       inPack.getLength(),
					       seq_num);
		if(killFlag)
		{
		    msgTable.removeElement(udpMessage);
		}
	    }

	}
    }
    /* handle_messages() */

    private
    UdpMessage
    findNode(
	UdpPacket				packet,
	int					seq_num)
    {
	int count;
	UdpMessage tempMsg = null;
	boolean found = false;

	InetAddress address;
	int         port;
	long        endpointID;

	address = packet.getAddress();
	port = packet.getPort();
	endpointID = packet.getEndpointID();

	count = msgTable.size();

	for(int ctr = 0; !found && ctr < count; ctr++)
	{
	    try
	    {
		tempMsg = (UdpMessage)(msgTable.elementAt(ctr));
	    }
	    catch(Exception e)
	    {
		System.out.println("element not found");
	    }

	    if(tempMsg.fromAddr.equals(address)
	       && tempMsg.fromPort == port
	       && tempMsg.endpointID == endpointID)
	    {
		found = true;
	    }
	}

	if(found)
	{
	    return tempMsg;
	}

	return null;
    }
    /* findNode() */

}
/* UdpHandler */


