/*
 * MulticastSingleRsr.java
 *
 * Doug Sale (ANL.DSL/NIU.HPC) - June 98
 */

import nexus.*;
import java.net.*;

public class MulticastSingleRsr {

  // usage and help strings
  public static final String USAGE = "java MulticastSingleRsr [-m] [-n <# nodes>] [-a <multicast IP address>] [-p <port #>] [-t <TTL>]";

  // Default Multicast Address, Port, Node TTL
  public static final String MULTICAST_ADDRESS = "239.255.255.255";
  public static final int PORT = 37428;
  public static final byte TTL = UdpInfo.RESTRICT_TO_SUBNET;

  // Handler ID's
  public static final int RECEIVE_RSR  = 1;

  // Nexus vars
  private Nexus        nexus;

  // Application vars
  private boolean master = false;
  private boolean nodes = false;
  private int     localNodes = 0;
  private int     received = 0;
  private String  address = null;
  private int     port = -1;
  private byte    ttl = -1;

  /*
   * MulticastSingleRsr()
   */
  public MulticastSingleRsr(String[] arg) {

    nexus = new Nexus();

    this.parseArgs(arg);
    this.runTest();

    nexus.destroy_current_context(false);
    System.exit(0);

  } /* MulticastSingleRsr */

  /*
   * runTest()
   */
  public void runTest() {

    if (nodes) {
      System.out.println("MulticastSingleRsr.runTest(): "
                   + "instantiating " + localNodes + " local nodes");
      for (int i = 0; i < localNodes; i++) {
        new Node(nexus, this, address, port);
      }
    }

    if (master) {
      System.out.println("MulticastSingleRsr.runTest(): "
                   + "instantiating master node");
      new Master(nexus, this, address, port, ttl).poll();
    }

    this.waitForRSRs();

  } /* MulticastSingleRsr() */

  /*
   * waitForRSRs()
   */
  public synchronized void waitForRSRs() {

    if (master) {
      localNodes++;
    }

    try {
      while (received != localNodes) {
        wait();
      }
    }
    catch (InterruptedException intX) {
      intX.printStackTrace();
    }

  } /* waitForRSRs() */

  /*
   * notifyOfReceipt()
   */
  public synchronized void notifyOfReceipt() {

    received++;
    notify();

  } /* notifyOfReceipt() */

  /*
   * parseArgs()
   */
  public void parseArgs(String[] arg) {

    int ndx = 0;

    while (ndx < arg.length) {
      if (arg[ndx].equals("-m")) {
        master = true;
        ndx++;
      }
      else if (arg[ndx].equals("-n")) {
        nodes = true;
        ndx++;

        try {
          localNodes = new Integer(arg[ndx]).intValue(); 
          ndx++;
        }
        catch (NumberFormatException numX) {
          System.err.println("MulticastSingleRsr: Invalid value ("
                             + arg[ndx] + ") for number of nodes");
          System.err.println(USAGE);
          nexus.destroy_current_context(false);
          System.exit(1);
        } 
        catch (ArrayIndexOutOfBoundsException ndxX) {
          System.err.println("MulticastSingleRsr: When using the nodes "
                             + "switch (-n) a number must be specified");
          System.err.println(USAGE);
          nexus.destroy_current_context(false);
          System.exit(1);
        }
      }
      else if (arg[ndx].equals("-a")) {
        ndx++;

        try {
          InetAddress addr = InetAddress.getByName(arg[ndx]);
          if (!addr.isMulticastAddress()) {
            throw new UnknownHostException(arg[ndx]);
          }
          address = arg[ndx];
          ndx++;
        }
        catch (UnknownHostException hostX) {
          System.err.println("MulticastSingleRsr: Invalid value ("
                             + arg[ndx] + ") for multicast IP address "
                             + "(valid range 224.0.1.0 - 239.255.255.255)"); 
          System.err.println(USAGE);
          nexus.destroy_current_context(false);
          System.exit(1);
        } 
        catch (ArrayIndexOutOfBoundsException ndxX) {
          System.err.println("MulticastSingleRsr: When using the address "
                             + "switch (-p) a multicast IP address "
                             + "must be specified");
          System.err.println(USAGE);
          nexus.destroy_current_context(false);
          System.exit(1);
        }
      }
      else if (arg[ndx].equals("-p")) {
        ndx++;

        try {
          port = new Integer(arg[ndx]).intValue(); 
          ndx++;
        }
        catch (NumberFormatException numX) {
          System.err.println("MulticastSingleRsr: Invalid value ("
                             + arg[ndx] + ") for port number");
          System.err.println(USAGE);
          nexus.destroy_current_context(false);
          System.exit(1);
        } 
        catch (ArrayIndexOutOfBoundsException ndxX) {
          System.err.println("MulticastSingleRsr: When using the port "
                             + "switch (-p) a number must be specified");
          System.err.println(USAGE);
          nexus.destroy_current_context(false);
          System.exit(1);
        }
      }
      else if (arg[ndx].equals("-t")) {
        ndx++;

        try {
          ttl = new Byte(arg[ndx]).byteValue(); 
          ndx++;
        }
        catch (NumberFormatException numX) {
          System.err.println("MulticastSingleRsr: Invalid value ("
                             + arg[ndx] + ") for TTL");
          System.err.println(USAGE);
          nexus.destroy_current_context(false);
          System.exit(1);
        } 
        catch (ArrayIndexOutOfBoundsException ndxX) {
          System.err.println("MulticastSingleRsr: When using the TTL "
                             + "switch (-t) a number must be specified");
          System.err.println(USAGE);
          nexus.destroy_current_context(false);
          System.exit(1);
        }
      }
    }

    if (!master && !nodes) {
      System.err.println(USAGE);
      nexus.destroy_current_context(false);
      System.exit(1);
    }

    if (address == null) {
      this.address = MulticastSingleRsr.MULTICAST_ADDRESS;
    }
    if (port == -1) {
      this.port = MulticastSingleRsr.PORT;
    }
    if (ttl == -1) {
      this.ttl = MulticastSingleRsr.TTL;
    }

  } /* parseArgs() */

  /*
   * main()
   */
  public static void main(String[] arg) {

    new MulticastSingleRsr(arg);
    
  } /* main() */

} /* MulticastSingleRsr */


/*
 * Master
 */
class Master implements HandlerInterface {

  // Application vars
  MulticastSingleRsr main;
  String address;
  int    port;
  byte   ttl;

  // Nexus vars
  private Nexus        nexus;
  private EndpointAttr epAttr;
  private Endpoint     ep;
  private Startpoint   sp;

  /*
   * Master()
   */
  public Master(Nexus nexus, MulticastSingleRsr main,
                String address, int port, byte ttl) {

    this.nexus = nexus;
    this.main = main;
    this.address = address;
    this.port = port;
    this.ttl = ttl;

    // init endpoint attribute
    epAttr = nexus.endpointattr_init();
    epAttr.set_protocol(new UdpInfo(address, port, ttl));
    epAttr.set_handler_object(this);

    // initialize endpoint 
    ep = nexus.endpoint_init(epAttr); 

    System.out.println("Master.Master(): listening on " + address + ":" + port);

  } /* Master() */

  /*
   * poll()
   */
  public void poll() {

    PutBuffer buffer;

    // bind startpoint
    sp = nexus.startpoint_bind(ep);

    System.out.println("Master.poll(): multicasting to RECEIVE_RSR on "
                       + address + ":" + port);

    // init buffer and send remote service request
    buffer = nexus.buffer_init(0, 0);
    nexus.send_rsr(buffer, sp, MulticastSingleRsr.RECEIVE_RSR, false);
  
  } /* poll() */

  /*
   * invoke_handler()
   * (HandlerInterface)
   */
  public void invoke_handler(Endpoint ep, GetBuffer buffer,
                             int handlerID) {
    switch(handlerID) {
      case MulticastSingleRsr.RECEIVE_RSR:
        System.out.println("Master.invoke_handler(): RECEIVE_RSR");
        main.notifyOfReceipt();
        break;
      default:
        System.err.println("Master.invoke_handler(): "
                           + "unknown handler id " + handlerID);
    }
  } /* invoke_handler() */

} /* Master */


/*
 * Node
 */
class Node implements HandlerInterface {

  // Application vars
  MulticastSingleRsr main;

  // Nexus vars
  private Nexus        nexus;
  private EndpointAttr epAttr;
  private Endpoint     ep;

  /*
   * Node()
   */
  public Node(Nexus nexus, MulticastSingleRsr main,
              String address, int port) {

    this.nexus = nexus;
    this.main = main;

    // init endpoint attribute
    epAttr = nexus.endpointattr_init();
    epAttr.set_protocol(new UdpInfo(address, port, UdpInfo.RESTRICT_TO_HOST));
    epAttr.set_handler_object(this);

    // initialize endpoint and bind startpoint
    ep = nexus.endpoint_init(epAttr); 

    System.out.println("Node.Node(): listening on " + address + ":" + port);

  } /* Node() */

  /*
   * invoke_handler()
   * (HandlerInterface)
   */
  public void invoke_handler(Endpoint ep,
                             GetBuffer buffer,
                             int handlerID) {
    switch(handlerID) {
      case MulticastSingleRsr.RECEIVE_RSR:
        System.out.println("Node.invokeHandler(): RECEIVE_RSR");
        main.notifyOfReceipt();
        break;
      default:
        System.err.println("Node.invoke_handler(): "
                           + "unknown handler id " + handlerID);
    }
  } /* invoke_handler() */

} /* Node */
