#include <stdio.h>
#include <string.h>
#include <globus_nexus.h>

/*
 * nexus handler functions
 */
static void dummy_handler(nexus_endpoint_t *endpoint,
                            nexus_buffer_t *buffer,
                            nexus_bool_t is_non_threaded_handler); 
static void reply_handler(nexus_endpoint_t *endpoint,
                            nexus_buffer_t *buffer,
                            nexus_bool_t is_non_threaded_handler); 
/*
 * nexus handler functions table
 */
#define RECEIVE_HANDLER		0
#define REPLY_HANDLER		1
#define SERVER_SHUTDOWN_HANDLER 2

static nexus_handler_t handlers[] = 
{ \

    {NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) dummy_handler},

    {NEXUS_HANDLER_TYPE_NON_THREADED,
       (nexus_handler_func_t) reply_handler}
};

/* Monitor                                                 */
/*                                                         */
/* we can simulate a monitor by defining a struct that has */
/* at least a nexus_mutex_t to provide mutually exclusive  */
/* access to the monitor.                                  */
/* in this application we also need a monitor condition    */
/* variable and a flag.                                    */

typedef struct
{
    /* for mutually exclusive access to monitor */
    nexus_mutex_t mutex;

    /* a monitor condition variable and flag */
    nexus_cond_t cond;
    nexus_bool_t done;
} monitor_t;

/* Global Variables */
static nexus_endpointattr_t	EpAttr;
static nexus_endpoint_t		GlobalEndpoint;
static nexus_startpoint_t	ServerSP;
static monitor_t		Monitor;

/* Local Functions */
static void attach_to_server(char *attach_url);

  extern int _nx_debug_level;

/*
 * main()
 */
int main(int argc, char **argv)
{
  nexus_startpoint_t this_sp;
  nexus_buffer_t buffer;
  char attach_url[1024];
  char c;
  short s;
  int i;
  float f;
  double d;  

  globus_module_activate(GLOBUS_NEXUS_MODULE);


  _nx_debug_level = 0;

  /* Initialization */
  nexus_endpointattr_init(&EpAttr);
  nexus_endpointattr_set_handler_table(&EpAttr, 
    handlers, sizeof(handlers)/sizeof(nexus_handler_t));

#if 0
  nexus_endpointattr_set_protocol(&EpAttr,
                                  NEXUS_PROTO_TYPE_UDP,
                                  NULL, 0); 
#endif


  /* initializing endpoint ... no particular address */
  nexus_endpoint_init(&GlobalEndpoint, &EpAttr);

  /* bind sp to this context */
  nexus_startpoint_bind(&this_sp, &GlobalEndpoint);

  c = 7;
  s = 24;
  i = 42;
  f = 12.345;
  d = 67.89;

  /* parse args */
  if (!(argc > 1 && argc < 3))
  {
    globus_fatal("usage: %s attach_url [-leave_server]\n", argv[0]);
  }
  else
  {
    strcpy(attach_url, argv[1]);
  } /* endif */

  /* initialize monitor */
  nexus_mutex_init(&(Monitor.mutex), (nexus_mutexattr_t *) NULL);
  nexus_cond_init(&(Monitor.cond), (nexus_condattr_t *) NULL);
  nexus_mutex_lock(&(Monitor.mutex));
  Monitor.done = NEXUS_FALSE;
  nexus_mutex_unlock(&(Monitor.mutex));

  /* attach to server */ 
  attach_to_server(attach_url);
  nexus_printf("test_client: attached to server\n");

  /* init and pack the buffer */
  nexus_buffer_init(&buffer, 
    nexus_sizeof_startpoint(&this_sp, 1)
    + nexus_sizeof_char(1)
    + nexus_sizeof_short(1)
    + nexus_sizeof_int(1)
    + nexus_sizeof_float(1)
    + nexus_sizeof_double(1),
    0);
  nexus_put_startpoint_transfer(&buffer, &this_sp, 1);
  nexus_put_char(&buffer, &c, 1);
  nexus_put_short(&buffer, &s, 1);
  nexus_put_int(&buffer, &i, 1);
  nexus_put_float(&buffer, &f, 1);
  nexus_put_double(&buffer, &d, 1);

  nexus_printf("test_client: packing buffer - this sp, c=%d, s=%d, i=%d, f=%f, d=%f\n", c, s, i, f, d);
  nexus_printf("test_client: send_rsr - RECEIVE_HANDLER\n");

  /* send rsr */
  nexus_send_rsr(&buffer, 
    &ServerSP, 
    RECEIVE_HANDLER, 
    NEXUS_TRUE, 
    NEXUS_FALSE);
  nexus_printf(" JB BS post send\n");

  /* waiting for reply */
  nexus_mutex_lock(&(Monitor.mutex));
  while (!(Monitor.done))
  {
    nexus_cond_wait(&(Monitor.cond), &(Monitor.mutex));
  } /* endwhile */
  nexus_mutex_unlock(&(Monitor.mutex));

  /* Clean-Up */
  nexus_startpoint_destroy(&ServerSP);
  nexus_mutex_destroy(&(Monitor.mutex));
  nexus_cond_destroy(&(Monitor.cond));
  nexus_endpoint_destroy(&GlobalEndpoint);
  nexus_endpointattr_destroy(&EpAttr);
  nexus_startpoint_destroy(&this_sp);
  
  globus_module_deactivate(GLOBUS_NEXUS_MODULE);
  
  exit(0);
} /* end main() */


/*
 * attach_to_server()
 */
static void attach_to_server(char *attach_url)
{
  int rc;

  rc = nexus_attach(attach_url, &ServerSP);

  if (rc != 0)
  {
    globus_fatal("test_client: Could not attach to server using url: "
		 "%s .  rc = %d\n", attach_url, rc);
  } /* endif */

  nexus_printf("test_client: Attached to %s\n", attach_url);
} /* end attach_to_server() */

/*
 * dummy_handler()
 */
static void dummy_handler(nexus_endpoint_t *endpoint, 
                          nexus_buffer_t *buffer,
                          nexus_bool_t is_non_threaded_handler)
{
  /* clean-up */
  nexus_buffer_destroy(buffer);
} /* dummy_handler() */

/*
 * reply_handler()
 */
static void reply_handler(nexus_endpoint_t *endpoint, 
			  nexus_buffer_t *buffer,
			  nexus_bool_t is_non_threaded_handler)
{
  nexus_printf("test_client: reply_handler entering\n");

  /* clean-up */
  nexus_buffer_destroy(buffer);

  /* init and pack the buffer */
  nexus_buffer_reset(buffer, 0, 0);
     
  /* send rsr */
  nexus_printf("test_client: send_rsr - SERVER_SHUTDOWN_HANDLER\n"); 
  nexus_send_rsr(buffer, 
                 &ServerSP, 
                 SERVER_SHUTDOWN_HANDLER, 
                 NEXUS_TRUE, 
                 NEXUS_FALSE);
  nexus_printf("test_client: send_rsr returned\n");  

  nexus_mutex_lock(&(Monitor.mutex));
  Monitor.done = NEXUS_TRUE;
  nexus_cond_signal(&(Monitor.cond));
  nexus_mutex_unlock(&(Monitor.mutex));   

  nexus_printf("test_client: reply_handler exiting\n");

} /* reply_handler() */

 
